"""PHY Injector connector
"""
import logging

from whad.phy.connector import Phy
from whad.hub.phy import Modulation as PhyModulation, Endianness as PhyEndianness
from whad.phy.injecting import InjectionConfiguration
from whad.scapy.layers.phy import Phy_Packet
from whad.exceptions import UnsupportedCapability

logger = logging.getLogger(__name__)

# TODO: every sniffer is broken (sniff() method does not catch packets, we
#       have to catch them in on_packet() and put them in a queue)

class Injector(Phy):
    """
    Phy Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        Phy.__init__(self, device)
        self._metadata_check = False
        # Check if device can perform injection
        if not self.can_send():
            raise UnsupportedCapability("Inject")

        self.__configuration = InjectionConfiguration()


    @property
    def configuration(self) -> InjectionConfiguration:
        """Retrieve this injector configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: InjectionConfiguration):
        """Set the injector configuration.
        """
        self.stop()
        self.__configuration = new_configuration
        if self.__configuration.frequency is not None:
            self.set_frequency(self.__configuration.frequency)
        if self.__configuration.packet_size is not None:
            if not self.set_packet_size(self.__configuration.packet_size):
                logger.warning(("[WARNING] Hardware rejected the specified packet size (%d), "
                                "will use its maximum packet size instead."),
                               self.__configuration.packet_size)

        # Set data rate for all modulations but LoRa
        if not self.__configuration.lora:
            if self.__configuration.datarate is not None:
                self.set_datarate(self.__configuration.datarate)

        if self.__configuration.gfsk:
            self.set_gfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.bfsk:
            self.set_bfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.qfsk:
            self.set_4fsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.ask:
            self.set_ask()
        elif self.__configuration.bpsk:
            self.set_bpsk()
        elif self.__configuration.qpsk:
            self.set_qpsk()
        elif self.__configuration.lora:
            self.set_lora(
                self.__configuration.lora_configuration.spreading_factor,
                self.__configuration.lora_configuration.coding_rate,
                self.__configuration.lora_configuration.bandwidth,
                self.__configuration.lora_configuration.preamble_length,
                crc=self.__configuration.lora_configuration.enable_crc,
                explicit=self.__configuration.lora_configuration.enable_explicit_mode
            )

        # Set endianness for all modulations but LoRa
        if not self.__configuration.lora:
            self.set_endianness(
                Endianness.LITTLE if
                self.__configuration.little_endian else
                Endianness.BIG
            )
        if self.__configuration.sync_word is not None:
            self.set_sync_word(self.__configuration.sync_word)
        self.start()

    def inject(self, packet):
        """Send packet through hardware.
        """
        if not self._metadata_check:
            if hasattr(packet, "metadata") and hasattr(packet.metadata, "modulation") and packet.metadata.modulation is not None:
                deviation = packet.metadata.deviation
            else:
                deviation = None
            if deviation is not None:
                self.stop()
                if hasattr(packet, "metadata") and hasattr(packet.metadata, "modulation"):
                    if packet.metadata.modulation == PhyModulation.ASK and not self.__configuration.ask:
                        self.set_ask()
                    elif packet.metadata.modulation == PhyModulation.FSK and deviation is not None and not self.__configuration.bfsk:
                        self.set_bfsk(deviation=deviation)
                    elif (packet.metadata.modulation == PhyModulation.GFSK or packet.metadata.modulation == PhyModulation.MSK) and deviation is not None and not self.__configuration.gfsk:
                        self.set_gfsk(deviation=deviation)
                    elif packet.metadata.modulation == PhyModulation.FOURFSK and deviation is not None and not self.__configuration.qfsk:
                        self.set_4fsk(deviation=deviation)
                    elif packet.metadata.modulation == PhyModulation.BPSK and not self.__configuration.bpsk:
                        self.set_bpsk()
                    elif packet.metadata.modulation == PhyModulation.QPSK and not self.__configuration.qpsk:
                        self.set_qpsk()
                    elif packet.metadata.modulation == PhyModulation.LORA and not self.__configuration.lora:
                        self.set_lora(
                            self.__configuration.lora_configuration.spreading_factor,
                            self.__configuration.lora_configuration.coding_rate,
                            self.__configuration.lora_configuration.bandwidth,
                            12,
                            crc=self.__configuration.lora_configuration.enable_crc,
                            explicit=self.__configuration.lora_configuration.enable_explicit_mode
                        )

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "frequency") and self.__configuration.frequency is None:
                    self.set_frequency(packet.metadata.frequency)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "packet_size") and self.__configuration.packet_size is None:
                    self.set_packet_size(packet.metadata.packet_size)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "datarate") and self.__configuration.datarate is None:
                    self.set_datarate(packet.metadata.datarate)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "syncword") and self.__configuration.sync_word is None:
                    self.set_sync_word(packet.metadata.syncword)

                if hasattr(packet, "metadata") and hasattr(packet.metadata, "endianness"):
                    self.set_endianness(
                        Endianness.LITTLE if
                        packet.metadata.endianness == PhyEndianness.LITTLE else
                        Endianness.BIG
                    )

                self.start()
                self._metadata_check = True

        # Send packet payload
        super().send(packet[Phy_Packet])



class EDMeasurement:
    """
    Maximum measurement of an energy detection scan.
    """
    def __init__(self, samples, channel_page, channel):
        self.max_sample = max(samples)
        self.channel_number = channel
        self.channel_page = channel_page

    def __repr__(self):
        return ("EDMeasurement("+
                "max_sample=" + str(self.max_sample) +", "
                "channel_page=" + str(self.channel_page)+", "
                "channel_number=" + str(self.channel_number) +
            ")"
        )



"""
Exceptions linked to the ZigBee Cluster Library
"""

class ZCLAttributePermissionDenied(Exception):
    """
    Exception raised when permission to access an attribute is denied.
    """
    def __init__(self):
        super().__init__()

class ZCLAttributeNotFound(Exception):
    """
    Exception raised when an attribute is not found.
    """
    def __init__(self):
        super().__init__()

class ZCLCommandNotFound(Exception):
    """
    Exception raised when a command is not found.
    """
    def __init__(self):
        super().__init__()



"""LoRaWAN application module

This module provides a main class, `LWApplication` that can be used to implement LoRaWAN compatible applications
that will run on an emulated LoRaWAN gateway. Applications must inherit from this base class.

The `LWApplication provides some useful features:
- it supports over-the-air device activation and keeps track of devices that joined the network ;
- it also supports activated by personalization (ABP) devices as well ;
- it provides a set of callbacks that can be overriden by the application

Registered nodes state is kept in a JSON file named by default on the name of the application's EUI, and stored in
the working directory. This file is managed by the `LWNodeRegistry` class.
"""
import sys
import json
from os import unlink
from os.path import exists, isfile

# Including type hint Iterator based on python version
if sys.version_info[0] == 3 and sys.version_info[1] >= 9:
    from collections.abc import Iterator
else:
    from typing import Iterator

from whad.lorawan.exceptions import InvalidNodeRegistryError

import logging
logger = logging.getLogger(__name__)

class LWNode(object):
    """LoRaWAN node
    """

    def __init__(self, dev_eui, dev_addr=0, appskey=b'', nwkskey=b'', upcount=0, dncount=0):
        """Initialize a LoRaWAN node (device)

        :param dev_eui: Device extended unique identifier (in the following format: `00:11:22:33:44:55:66:77`)
        :type dev_eui: str
        :param dev_addr: Device address on the network
        :type dev_addr: int, optional
        :param appskey: Device application session key
        :type appskey: bytes, optional
        :param nwkskey: Device network session encryption key
        :type nwkskey: bytes, optional
        :param upcount: Device current uplink frame counter (default: 0)
        :type upcount: int, optional
        :param dncount: Device current downlink frame counter (default: 0)
        :type dncount: int, optional
        """
        self.__dev_eui = str(dev_eui).lower()
        self.__dev_addr = dev_addr
        self.__appskey = appskey
        self.__nwkskey = nwkskey
        self.__upcount = upcount
        self.__dncount = dncount

    def __repr__(self):
        return 'LWNode(eui:%s, address:%s, appskey:%s, nwkskey:%s, uplink_count:%d, downlink_count:%d)' % (
            self.dev_eui,
            self.dev_addr,
            self.__appskey.hex() if self.__appskey is not None else None,
            self.__nwkskey.hex() if self.__nwkskey is not None else None,
            self.__upcount,
            self.__dncount
        )

    @property
    def dev_eui(self) -> str:
        """Device EUI getter
        """
        return self.__dev_eui
    
    @property
    def dev_addr(self) -> int:
        """Device network address getter

        :return: Device address
        :rtype: int
        """
        return self.__dev_addr
    
    @property
    def appskey(self) -> bytes:
        """Device application session key getter

        :return: Device application session key
        :rtype: bytes
        """
        return self.__appskey
    
    @property
    def nwkskey(self) -> bytes:
        """Device network session encryption key

        :return: Device network session encryption key
        :rtype: bytes
        """
        return self.__nwkskey
    
    @property
    def upcount(self) -> int:
        """Uplink frame counter getter

        :return: Current device uplink frame counter
        :rtype: int
        """
        return self.__upcount
    
    @upcount.setter
    def upcount(self, value: int):
        """Uplink frame counter setter

        :param value: New uplink frame counter value for the device
        :type value: int
        """
        self.__upcount = value
    
    @property
    def dncount(self) -> int:
        """Downlink frame counter getter

        :return: Current device downlink frame counter
        :rtype: int
        """
        return self.__dncount

    @dncount.setter
    def dncount(self, value: int):
        """Downlink frame counter setter

        :param value: New downlink frame counter value for the device
        :type value: int
        """
        self.__upcount = value

    @property
    def joined(self) -> bool:
        """Check if the device has joined a network

        :return: `True` if the device has joined a network, `False` otherwise.
        :rtype: bool
        """
        return (self.appskey is not None and self.nwkskey is not None and self.dev_addr is not None)

    def inc_up(self):
        """Increment device's uplink frame counter
        """
        self.__upcount = (self.__upcount + 1) & 0xffffffff

    def inc_down(self):
        """Increment device's downlink frame counter
        """
        self.__dncount = (self.__dncount + 1) & 0xffffffff

    def toDict(self) -> dict:
        """Return this node as a dictionary

        :return: Device properties as a dictionary
        :rtype: dict
        """
        return {
            'dev_eui': self.dev_eui,
            'dev_addr': self.dev_addr,
            'appskey': self.appskey.hex(),
            'nwkskey': self.nwkskey.hex(),
            'upcount': self.upcount,
            'dncount': self.dncount
        }
    
    @staticmethod
    def fromJSON(data):
        """Create a `LWNode` instance from a node saved state (JSON).

        :param data: JSON data representing the device state
        :type data: str
        :return: An instance of `LWNode` corresponding to the serialized device state
        :rtype: LWNode
        """
        return LWNode(
            data["dev_eui"],
            data["dev_addr"],
            bytes.fromhex(data["appskey"]),
            bytes.fromhex(data["nwkskey"]),
            data["upcount"],
            data["dncount"]
        )
    

class LWNodeRegistry(object):
    """LoRaWAN node registry.

    This class holds a list of allowed devices EUI and
    associated addresses and encryption keys for persistent
    purpose.

    Data is stored in a flat json file.
    """

    def __init__(self, path:str ='default_node.json'):
        """Load node registry file.

        :param path: Default registry file
        :type path: str
        """
        self.__path = path
        self.__nodes = {}
        if exists(path) and isfile(path):
            try:
                # File exists, load it as json
                with open(path, 'r') as registry:
                    nodes = json.load(registry)

                    # Loop on nodes and load our lookup table
                    for node in nodes:
                        node_ = LWNode.fromJSON(node)
                        self.__nodes[node_.dev_eui] = node_
                    registry.close()
            except IOError:
                raise InvalidNodeRegistryError(path)
            except Exception:
                # Oops, error while loading registry. Unlink file.
                logger.error("Error while loading registry file %s, removing file.", path)
                unlink(path)
        else:
            # File does not exist, create it.
            try:
                self.save()
            except IOError:
                raise InvalidNodeRegistryError(path)

    def add_node(self, node: LWNode):
        """Register a LoRaWAN node.

        :param node: Node to add to this registry
        :type node: LWNode
        """
        if node.dev_eui in self.__nodes:
            logger.warning("Device %s is already present in node registry", node.dev_eui)
            self.__nodes[node.dev_eui] = node
        else:
            logger.debug("adding node %s", node)
            self.__nodes[node.dev_eui] = node

    def get_node(self, eui:str = None) -> LWNode:
        """Retrieve node by DEV EUI.

        :param eui: Node EUI
        :type eui: str
        :return: Corresponding device instance if found, `None` if not.
        :rtype: LWNode
        """
        if str(eui) in self.__nodes:
            return self.__nodes[str(eui)]
        else:
            return None

    def iterate(self) -> Iterator[LWNode]:
        """Iterate over registered nodes.
        """
        for dev_eui in self.__nodes:
            yield self.__nodes[dev_eui]

    def save(self):
        """Save registry to file
        """
        try:
            with open(self.__path, 'w') as registry:
                nodes = []
                for node_eui in self.__nodes:
                    nodes.append(self.__nodes[node_eui].toDict())
                json.dump(nodes, registry)
                registry.close()
        except IOError:
            raise InvalidNodeRegistryError(self.__path)

    
class LWApplication(object):
    """LoRaWAN application class
    """

    def __init__(self, eui=None, key=None, node_db_path=None, devices:[LWNode] = []):
        """Initialize a LoRaWAN application

        :param eui: Application EUI
        :type eui: str
        :param key: Application key in hexadecimal form
        :type key: str
        :param node_db_path: Application database path, default is named <APP_EUI>.json
        :type node_db_path: str
        :param devices: Allowed devices
        :type devices: list
        """
        # Save application EUI and main key
        self.__eui = str(eui).lower()
        self.__key = key

        # Initialize node registry
        if node_db_path is not None:
            self.__registry = LWNodeRegistry(path=node_db_path)
        else:
            self.__registry = LWNodeRegistry(
                path='%s.json' % self.__eui.replace(':','')
            )

        # Load registry with devices
        for device in devices:
            self.__registry.add_node(device)

        # Initialize downlink data buffer
        self.__pending_data = b''

    @property
    def eui(self) -> str:
        """Application EUI getter

        :return: Application EUI
        :rtype: str
        """
        return self.__eui
    
    @property
    def key(self) -> str:
        """Application key getter

        :return: Application key as an hexadecimal string
        :rtype: str
        """
        return self.__key
    
    def nodes(self) -> Iterator[LWNode]:
        """Iterate over registered nodes.
        """
        for node in self.__registry.iterate():
            yield node

    def add_node(self, node: LWNode = None):
        """Dynamically add a node to application registry.

        :param node: Node to add to this application
        :type node: LWNode
        """
        if node is not None:
            self.__registry.add_node(node)

    def stop(self):
        """Stop this application.

        Stopping the application will save the current node registry status into the application state JSON file.
        """
        # Save node registry.
        self.__registry.save()

    def is_authorized(self, dev_eui) -> bool:
        """Determine if device is authorized to join the network.

        This method can be overriden to implement additional checks on device.

        :param dev_eui: Device EUI
        :type dev_eui: EUI
        :return: True if device is authorized, False otherwise
        :rtype: bool
        """
        # Basically, if we have a node registered for this device that's ok.
        return (self.__registry.get_node(dev_eui) is not None)

    def on_device_joined(self, dev_eui, dev_addr, appskey, nwkskey):
        """Handles device join procedure

        This method is called whenever a device has joined the network through OTAA.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param appskey: Device's application session key
        :type appskey: bytes
        :param nwkskey: Device's network session encryption key
        :type nwkskey: int
        """
        # Add this device to our registry
        self.__registry.add_node(LWNode(
            dev_eui,
            dev_addr,
            appskey,
            nwkskey
        ))

    def on_device_data(self, dev_eui, dev_addr, data:bytes, upcount=0):
        """Handles data sent by a device.

        :param dev_eui: Device EUI
        :type dev_eui: EUI
        :param dev_addr: Device address
        :type dev_addr: int
        :param data: Data sent by device
        :type data: bytes
        :param upcount: uplink frame counter
        :type upcount: int
        """
        node = self.__registry.get_node(dev_eui)
        if node is not None:
            # Update uplink frame counter
            node.upcount = upcount

            # Make sure device has joined
            if node.joined:
                # We ask our specific callback
                return self.on_data(node, data)
            else:
                # If not, we should not have received somthing
                return None
        else:
            return None


    def on_data(self, node: LWNode, data: bytes) -> bytes:
        """This callback handles data coming from a device.

        If it returns some bytes, these bytes will be sent back
        to the device through a downlink transmission.

        If None is returned, nothing is sent back to the device.

        :param node: Node that sent the data
        :type node: LWNode
        :param data: Data sent by the node
        :type data: bytes

        :return: Data to send back to the device
        :rtype: bytes
        """
        return None


"""WHAD Protocol BLE hijacking messages abstraction layer.
"""

from whad.hub.message import pb_bind, PbFieldInt, PbMessageWrapper, PbFieldBool
from whad.hub.ble import BleDomain

@pb_bind(BleDomain, "hijack_master", 1)
class HijackMaster(PbMessageWrapper):
    """BLE hijack master message class
    """
    access_address = PbFieldInt("ble.hijack_master.access_address")

@pb_bind(BleDomain, "hijack_slave", 1)
class HijackSlave(PbMessageWrapper):
    """BLE hijack slave message class
    """
    access_address = PbFieldInt("ble.hijack_slave.access_address")

@pb_bind(BleDomain, "hijack_both", 1)
class HijackBoth(PbMessageWrapper):
    """BLE hijack both message class
    """
    access_address = PbFieldInt("ble.hijack_both.access_address")

@pb_bind(BleDomain, "hijacked", 1)
class Hijacked(PbMessageWrapper):
    """BLE hijack both message class.
    """
    success = PbFieldBool("ble.hijacked.success")
    access_address = PbFieldInt("ble.hijacked.access_address")



from whad.unifying.connector import Sniffer
from whad.unifying.hid import LogitechUnifyingKeystrokeConverter, HIDCodeNotFound, InvalidHIDData
from whad.scapy.layers.unifying import Logitech_Unencrypted_Keystroke_Payload, Logitech_Encrypted_Keystroke_Payload

class Keylogger(Sniffer):
    """
    Logitech Unifying Keylogger interface for compatible WHAD device.
    """

    def __init__(self, device):
        super().__init__(device)
        self.__locale = "fr"
        self.key = None

    @property
    def locale(self):
        """ Current locale """
        return self.__locale

    @locale.setter
    def locale(self, locale):
        """Update locale."""
        self.__locale = locale

    def stream(self):
        for packet in super().sniff():
            hid_data = None
            if Logitech_Unencrypted_Keystroke_Payload in packet:
                hid_data = packet.hid_data

            if Logitech_Encrypted_Keystroke_Payload in packet:
                hid_data = packet.hid_data

            if hid_data is not None:
                if hid_data == b"\x00" * 7:
                    self.key = None
                else:
                    try:
                        key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(hid_data, locale=self.__locale)

                        if key != self.key:
                            self.key = key
                            if len(self.key) > 1:
                                self.key = " [{}] ".format(self.key)
                            yield self.key
                    except (HIDCodeNotFound, InvalidHIDData):
                        pass



"""Helper functions to parse DarkMentorLLC's CLUES collaborative database.
"""
import json
import os.path
import logging

from importlib import resources

from whad.ble.profile.attribute import UUID

logger = logging.getLogger(__name__)

def uuid_match(uuid: UUID, pattern: str) -> bool:
    """Determine if the provided UUID matches the pattern.
    """
    # Convert UUID to str
    uuid_s = str(uuid)

    # Make sure both have same length
    if len(pattern) != len(uuid_s):
        return False

    for pos, c in enumerate(pattern):
        # 'x' in pattern is a wildcard
        if c == 'x':
            continue

        # If characters do not match, UUID is different
        if uuid_s[pos] != c:
            return False

    # Success
    return True


class CluesDb:
    """DarkMentorLLC Clues collaborative database.
    """
    CLUES_CACHE = []
    loaded = False

    @staticmethod
    def load_data() -> bool:
        """Load data from CLUES_data.json file
        """
        result = False

        if not CluesDb.loaded:
            # Load data from CLUES_data.json into cache
            clues_data_path = os.path.join(resources.files("whad"),
                                           "resources/clues/CLUES_data.json")

            # Ensure the database file is present
            if os.path.exists(clues_data_path):
                try:
                    with open(clues_data_path, 'r', encoding="utf-8") as clues_json:
                        CluesDb.CLUES_CACHE = json.load(clues_json)

                    # Success
                    result = True
                except IOError:
                    logger.debug("[cluesdb] input/output error while trying to open file %s", clues_data_path)
                    logger.error("CLUES database could not be loaded (read error)")
            else:
                logger.debug("[cluesdb] missing database file CLUES_data.json, expected path: %s", clues_data_path)
                logger.error("CLUES database could not be be found (missing file)")

        # Mark DB as loaded, even if it failed (avoiding multiple error messages).
        CluesDb.loaded = True

        # Return operation result
        return result

    @staticmethod
    def get_uuid_alias(uuid: UUID) -> str:
        """Generate alias based on UUID.
        """
        # Load data into cache
        CluesDb.load_data()

        # Loop on services and known UUIDS
        for clue in CluesDb.CLUES_CACHE:
            # Is it our UUID ?
            if uuid_match(uuid, clue["UUID"]):
                # If UUID belongs to a service, include company name
                if "GATT Service" in clue["UUID_usage_array"] and "UUID_name" in clue:
                    name = clue["UUID_name"]
                    company = clue["company"]
                    return f"{company} | {name}"
                if "GATT Characteristic" in clue["UUID_usage_array"] and "UUID_name" in clue:
                    name = clue["UUID_name"]
                    return f"{clue['UUID_name']}"

        # Not found
        return None



from whad.dot15d4.stack.database import Dot15d4Database

class MACPIB(Dot15d4Database):
    """
    802.15.4 MAC PIB Database of attributes.
    """

    def reset(self):
        """
        Reset the PIB database to its default value.
        """
        self.macExtendedAddress = 0x1122334455667788
        self.macAssociatedPanCoord = False
        self.macAssociationPermit = True
        self.macAutoRequest = False
        self.macDataSequenceNumber = 0
        self.macBeaconSequenceNumber = 0
        self.macBeaconOrder = 15
        self.macBeaconPayload = b""
        self.macBeaconAutoRespond = True
        self.macBattLifeExt = 15
        self.macSuperframeOrder = 15
        self.macRxOnWhenIdle = True
        self.macPanId = 0xFFFF
        self.macShortAddress = 0xFFFF
        self.macCoordShortAddress = 0
        self.macCoordExtendedAddress = 0
        self.macPromiscuousMode = False
        self.macImplicitBroadcast = False
        self.macResponseWaitTime = 32

        self.macLastBeacon = None

        self.macAckTimeout = 0.3



"""Mock for WHAD default connector.

This module provides a mock connector that implement every required method.
"""
from whad.hub.message import HubMessage

from ..connector import Connector
from ..device import Device

class MockConnector(Connector):

    def __init__(self, device: Device = None):
        super().__init__(device)
        self.__discovery = []
        self.__generic = []
        self.__domain = []
        self.__packets = []
        self.__events = []

    def on_discovery_msg(self, message: HubMessage):
        """Handle discovery messages.

        :param message: Discovery message
        :type message: HubMessage
        """
        self.__discovery.append(message)

    def on_generic_msg(self, message: HubMessage):
        """Handle generic messages.

        :param message: Generic message
        :type message: HubMessage
        """
        self.__generic.append(message)

    def on_domain_msg(self, message: HubMessage):
        """Handle domain messages.

        :param message: Domain message
        :type message: HubMessage
        """
        self.__domain.append(message)

    def on_packet(self, packet):
        """Default handler for received packet.

        :param packet: Received packet
        :type packet: Packet
        """
        self.__packets.append(packet)

    def on_event(self, event):
        """Default handler for received event.

        :param packet: Received event
        :type packet: Event
        """
        self.__events.append(event)




from scapy.packet import Packet, bind_layers
from scapy.fields import XStrField, IntField, SignedIntField, IntEnumField, FieldLenField, XStrLenField
from scapy.config import conf

MODULATIONS = {
    0 : "ASK",
    1 : "FSK",
    2 : "4FSK",
    3 : "GFSK",
    4 : "MSK",
    5 : "BFSK",
    6 : "QFSK",
    7 : "LoRa"
}
class Phy_Packet_Hdr(Packet):
    name = "Physical layer generic packet header"
    fields_desc = [
        IntField("frequency", None),
        SignedIntField("rssi", None),
        IntField("datarate", None),
        IntField("deviation", None),
        IntEnumField("endianness", None, {0 : "little", 1 : "big"}),
        IntEnumField("modulation", None, MODULATIONS),
        FieldLenField("syncword_length", None, length_of="syncword"),
        XStrLenField("syncword", None, length_from=lambda s: s.syncword_length)
    ]

class Phy_Packet(Packet):
    name = "Physical layer generic Packet"
    fields_desc = [
        XStrField("data", None)
    ]

bind_layers(Phy_Packet_Hdr, Phy_Packet)
conf.l2types.register(152, Phy_Packet_Hdr)



from dataclasses import dataclass, field

@dataclass
class InjectionConfiguration:
    """
    Configuration for injecting into ZigBee communication.

    :param channel: select the channel to use for injection (c)

    """
    channel : int = None



from whad.zigbee.stack.nwk.constants import ZigbeeRelationship, ZigbeeDeviceType
from time import time
class ZigbeeNode:
    """
    Class representing a Zigbee device.
    """
    def __init__(
                    self,
                    address,
                    device_type,
                    extended_address=None,
                    rx_on_when_idle=False,
                    end_device_configuration=0,
                    timeout_counter=None,
                    device_timeout=None,
                    transmit_failure=0,
                    lqi=None,
                    outgoing_cost=0,
                    age=0,
                    keepalive_received=False,
                    extended_pan_id=None,
                    logical_channel=None,
                    depth=None,
                    beacon_order=None,
                    permit_joining=None,
                    potential_parent=None,
                    update_id=None,
                    pan_id=None,
                    relationship=ZigbeeRelationship.NONE
    ):
        self.address = address
        self.device_type = device_type
        self.extended_address = extended_address
        self.rx_on_when_idle = rx_on_when_idle
        self.end_device_configuration = end_device_configuration
        self.timeout_counter = timeout_counter
        self.device_timeout = device_timeout
        self.transmit_failure = transmit_failure
        self.lqi = lqi
        self.outgoing_cost = outgoing_cost
        self.age = age
        self.update_id=update_id
        self.keepalive_received = keepalive_received
        self.extended_pan_id = extended_pan_id
        self.logical_channel = logical_channel
        self.depth = depth
        self.beacon_order = beacon_order
        self.permit_joining = permit_joining
        self.potential_parent = potential_parent
        self.relationship = relationship
        self.pan_id = pan_id
        self.lqis = [lqi] if lqi is not None else []
        self.last_update = time()

    def link_cost(self):
        pl = sum(self.lqis) / len(self.lqis)
        result = min(7,round(1/(pl ** 4)))
        self.__dict__["outgoing_cost"] = result
        return result

    def __setattr__(self, name, value):
        if hasattr(self, "last_update"):
            if name == "lqi":
                self.__dict__["lqis"].append(value)
                if len(self.__dict__["lqis"]) > 20:
                    self.__dict__["lqis"] = self.__dict__["lqis"][1:]
                    self.link_cost()
            self.__dict__["last_update"] = time()

        self.__dict__[name] = value

    def __repr__(self):
        if self.device_type == ZigbeeDeviceType.END_DEVICE:
            role = "ZigbeeEndDevice"
        elif self.device_type == ZigbeeDeviceType.COORDINATOR:
            role = "ZigbeeCoordinator"
        elif self.device_type == ZigbeeDeviceType.ROUTER:
            role = "ZigbeeRouter"

        return "{}(address={:04x}, extended_pan_id={:04x} - last update {} seconds ago)".format(
                    role,
                    self.address,
                    self.extended_pan_id,
                    round(time() - self.last_update, 2)
        )

class ZigbeeEndDevice(ZigbeeNode):
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    rx_on_when_idle=False,
                    end_device_configuration=0,
                    timeout_counter=None,
                    device_timeout=None,
                    transmit_failure=0,
                    lqi=0,
                    outgoing_cost=0,
                    age=0,
                    keepalive_received=False,
                    extended_pan_id=None,
                    logical_channel=None,
                    depth=None,
                    beacon_order=None,
                    permit_joining=None,
                    potential_parent=None,
                    update_id=None,
                    pan_id=None,
                    relationship=ZigbeeRelationship.NONE
    ):
        super().__init__(
                            address,
                            device_type=ZigbeeDeviceType.END_DEVICE,
                            extended_address=extended_address,
                            rx_on_when_idle=rx_on_when_idle,
                            end_device_configuration=end_device_configuration,
                            timeout_counter=timeout_counter,
                            device_timeout=device_timeout,
                            transmit_failure=transmit_failure,
                            lqi=lqi,
                            outgoing_cost=outgoing_cost,
                            age=age,
                            keepalive_received=keepalive_received,
                            extended_pan_id=extended_pan_id,
                            logical_channel=logical_channel,
                            depth=depth,
                            beacon_order=beacon_order,
                            permit_joining=permit_joining,
                            potential_parent=potential_parent,
                            update_id=update_id,
                            pan_id=pan_id,
                            relationship=relationship
        )

class ZigbeeCoordinator(ZigbeeNode):
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    rx_on_when_idle=False,
                    end_device_configuration=0,
                    timeout_counter=None,
                    device_timeout=None,
                    transmit_failure=0,
                    lqi=0,
                    outgoing_cost=0,
                    age=0,
                    keepalive_received=False,
                    extended_pan_id=None,
                    logical_channel=None,
                    depth=None,
                    beacon_order=None,
                    permit_joining=None,
                    potential_parent=None,
                    update_id=None,
                    pan_id=None,
                    relationship=ZigbeeRelationship.NONE
    ):
        super().__init__(
                            address,
                            device_type=ZigbeeDeviceType.COORDINATOR,
                            extended_address=extended_address,
                            rx_on_when_idle=rx_on_when_idle,
                            end_device_configuration=end_device_configuration,
                            timeout_counter=timeout_counter,
                            device_timeout=device_timeout,
                            transmit_failure=transmit_failure,
                            lqi=lqi,
                            outgoing_cost=outgoing_cost,
                            age=age,
                            keepalive_received=keepalive_received,
                            extended_pan_id=extended_pan_id,
                            logical_channel=logical_channel,
                            depth=depth,
                            beacon_order=beacon_order,
                            permit_joining=permit_joining,
                            potential_parent=potential_parent,
                            update_id=update_id,
                            pan_id=pan_id,
                            relationship=relationship
        )

class ZigbeeRouter(ZigbeeNode):
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    rx_on_when_idle=False,
                    end_device_configuration=0,
                    timeout_counter=None,
                    device_timeout=None,
                    transmit_failure=0,
                    lqi=0,
                    outgoing_cost=0,
                    age=0,
                    keepalive_received=False,
                    extended_pan_id=None,
                    logical_channel=None,
                    depth=None,
                    beacon_order=None,
                    permit_joining=None,
                    potential_parent=None,
                    update_id=None,
                    pan_id=None,
                    relationship=ZigbeeRelationship.NONE
    ):
        super().__init__(
                            address,
                            device_type=ZigbeeDeviceType.ROUTER,
                            extended_address=extended_address,
                            rx_on_when_idle=rx_on_when_idle,
                            end_device_configuration=end_device_configuration,
                            timeout_counter=timeout_counter,
                            device_timeout=device_timeout,
                            transmit_failure=transmit_failure,
                            lqi=lqi,
                            outgoing_cost=outgoing_cost,
                            age=age,
                            keepalive_received=keepalive_received,
                            extended_pan_id=extended_pan_id,
                            logical_channel=logical_channel,
                            depth=depth,
                            beacon_order=beacon_order,
                            permit_joining=permit_joining,
                            potential_parent=potential_parent,
                            update_id=update_id,
                            pan_id=pan_id,
                            relationship=relationship
        )



"""WHAD Protocol ESB mode messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.unifying.unifying_pb2 import StartCmd, StopCmd, SniffPairingCmd
from ..message import pb_bind, PbFieldBytes, PbFieldBool, PbFieldInt, PbMessageWrapper
from . import UnifyingDomain

@pb_bind(UnifyingDomain, 'sniff', 1)
class SniffMode(PbMessageWrapper):
    """Logitech Unifying SniffMode message
    """

    channel = PbFieldInt('unifying.sniff.channel')
    address = PbFieldBytes('unifying.sniff.address')
    show_acks = PbFieldBool('unifying.sniff.show_acknowledgements')

@pb_bind(UnifyingDomain, 'jam', 1)
class JamMode(PbMessageWrapper):
    """Logitech Unifying JamMode message
    """

    channel = PbFieldInt('unifying.jam.channel')

@pb_bind(UnifyingDomain, 'jammed', 1)
class Jammed(PbMessageWrapper):
    """Logitech Unifying Jammed notification message
    """

    timestamp = PbFieldInt('unifying.jammed.timestamp')

@pb_bind(UnifyingDomain, 'start', 1)
class UnifyingStart(PbMessageWrapper):
    """Logitech Unifying UnifyingStart message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.unifying.start.CopyFrom(StartCmd())

@pb_bind(UnifyingDomain, 'stop', 1)
class UnifyingStop(PbMessageWrapper):
    """Logitech Unifying UnifyingStop message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.unifying.stop.CopyFrom(StopCmd())

@pb_bind(UnifyingDomain, 'dongle', 1)
class DongleMode(PbMessageWrapper):
    """Logitech Unifying DongleMode message
    """

    channel = PbFieldInt('unifying.dongle.channel')

@pb_bind(UnifyingDomain, 'keyboard', 1)
class KeyboardMode(PbMessageWrapper):
    """Logitech Unifying KeyboardMode message
    """

    channel = PbFieldInt('unifying.keyboard.channel')

@pb_bind(UnifyingDomain, 'mouse', 1)
class MouseMode(PbMessageWrapper):
    """Logitech Unifying MouseMode message
    """

    channel = PbFieldInt('unifying.mouse.channel')

@pb_bind(UnifyingDomain, 'sniff_pairing', 1)
class SniffPairing(PbMessageWrapper):
    """Logitech Unifying SniffPairing message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.unifying.sniff_pairing.CopyFrom(SniffPairingCmd())



"""BLE Channel Map helper
"""

from typing import List, Generator
from struct import unpack

class ChannelMap(object):
    """BLE Channel Map class
    """

    @staticmethod
    def from_int(channel_map: int):
        """Create a ChannelMap from its int representation

        :param channel_map: Channel map value
        :type channel_map: int
        :return: Channel map
        :rtype: ChannelMap
        """
        channels = []
        for i in range(40):
            if (channel_map & (1 << i)) > 0:
                channels.append(i)
        return ChannelMap(channels)

    @staticmethod
    def from_bytes(channel_map: bytes):
        """Create a ChannelMap from its bytes representation

        :param channel_map: Channel map value
        :type channel_map: bytes
        :return: Channel map
        :rtype: ChannelMap
        """
        assert len(channel_map) == 5
        chanmap_int = unpack("<I", channel_map[:4])[0]
        chanmap_int += channel_map[4] << 32
        return ChannelMap.from_int(chanmap_int)



    def __init__(self, channels: List[int] = None):
        """Initialize our channel map

        :param channels: List of channels to include in channel map
        :type channels: list, optional
        """
        self.__map = 0

        # Loop over channels and add them to our map
        if channels is not None:
            for channel in channels:
                
                # Add channel to our channel map
                self.add(channel)

    def add(self, channel: int):
        """Add channel to our channel map.

        :param channel: Channel number to add to channel map
        :param channel: int
        """
        # Check channel number validity
        if channel < 0 or channel > 37:
            raise ValueError()
        
        # Add channel to our map
        self.__map |= (1 << channel)

    def remove(self, channel: int):
        """Remove channel from our channel map.

        :param channel: Channel number to remove from channel map
        :param channel: int
        """
        # Check channel number validity
        if channel < 0 or channel > 37:
            raise ValueError()
        
        # Remove channel from map
        if self.has(channel):
            self.__map = self.__map & ((1 << channel) ^ 0xFFFFFFFFFF)

    def has(self, channel: int) -> bool:
        """Check if a given channel is present in channel map.

        :param channel: Channel to check
        :type channel: int
        :return: True if channel is present in channel map, False otherwise
        :rtype: bool
        """
        return (self.__map & (1 << channel)) != 0
        
    def channels(self) -> Generator:
        """Iterate over channels
        """
        for channel in range(38):
            if self.has(channel):
                yield channel

    @property
    def value(self):
        """Retrieve the channel map value
        """
        return self.__map.to_bytes(5, 'little', signed=False)

# Default channel map
DefaultChannelMap = ChannelMap(channels=range(38))


"""Bluetooth Low Energy Human Interface Device Service Profile

Still a work in progress, but looks cool :D
"""

from whad.ble.profile.attribute import UUID
from whad.ble.profile import PrimaryService, Characteristic
from whad.ble.profile.services import BatteryService, DeviceInformationService

class HidMouseService(BatteryService, DeviceInformationService):
    """Bluetooth Low Energy HID Mouse Service
    """

    hid = PrimaryService(
        uuid=UUID(0x1812),

        # Mandatory characteristics
        info=Characteristic(
            uuid=UUID(0x2A4A),
            permissions=['read'],
        ),
        control_point=Characteristic(
            uuid=UUID(0x2A4C),
            permissions=['write_without_response']
        ),
        report_map=Characteristic(
            uuid=UUID(0x2A4B),
            permissions=['read']
        ),

        # Mouse-specific characteristics

    )



"""BLE Interactive shell
"""

from prompt_toolkit import print_formatted_text, HTML
from whad.cli.app import command
from whad.zigbee.cli.enddevice.shell import ZigbeeEndDeviceShell

@command('interactive')
def interactive_handler(app, command_args):
    """interactive Zigbee shell

    <ansicyan><b>interactive</b></ansicyan>

    Starts an interactive shell and let you interact with a Zigbee WHAD network:
    - scan networks
    - join a network
    - list applications and clusters
    - trigger actions
    """
    # We need to have an interface specified
    if app.interface is not None:
        # Launch an interactive shell
        myshell = ZigbeeEndDeviceShell(app.interface)
        myshell.run()
    else:
        # If stdin is piped, that means previous program has failed.
        # We display this warning only if the tool has been launched in
        # standalone mode
        if not app.is_stdin_piped():
            app.error('You need to specify an interface with option --interface.')



from whad.dot15d4.stack.database import Dot15d4Database

class APLIB(Dot15d4Database):
    """
    RF4CE APLIB Database of attributes.
    """

    def reset(self):
        """
        Reset the APL database to its default value.
        """
        self.aplcMaxKeyRepeatInterval = 120 # (ms)
        self.aplcMaxRIBAttributeSize = 92
        self.aplcResponseIdleTime = 50 # (ms)
        self.aplcBlackOutTime = 100 # (ms)
        self.aplcMinKeyExchangeTransferCount = 3

        self.aplKeyRepeatInterval = 0.5 * self.aplcMaxKeyRepeatInterval
        self.aplKeyRepeatWaitTime = self.aplcMaxKeyRepeatInterval
        self.aplResponseWaitTime = 0x186a # (symbols ~ 16us, default = 100ms)
        self.aplMaxPairingCandidates = 3
        self.aplLinkLostWaitTime = 0x02710 # (symbols ~ 16us, default = 10000ms)
        self.aplAutoCheckValidationPeriod = 0x01f4 # (symbols ~ 16us, default = 500ms)
        self.aplValidationWaitTime = 0x0 # (infinity, or 0x7530 symbols / 30000 ms for button based validation)
        self.aplValidationInitialWatchdogTime = 0x1f40 # (symbols ~ 16us, default = 8000ms or 0 (infinity) for button based validation)
        self.aplUserString = "WHAD"
        self.aplKeyExchangeTransferCount = 4

        self.aplDeviceType = 9



import logging
from time import time
from typing import Generator

from scapy.packet import Packet
from scapy.layers.zigbee import ZigbeeSecurityHeader
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter
from whad.common.sniffing import EventsManager
from whad.hub.dot15d4 import RawPduReceived, PduReceived
from whad.hub.message import AbstractPacket
from whad.exceptions import WhadDeviceDisconnected

from whad.device import Device
from ..connector import Zigbee
from ..sniffing import SnifferConfiguration, KeyExtractedEvent
from ..crypto import ZigbeeDecryptor, TouchlinkKeyManager, TransportKeyDistribution


logger = logging.getLogger(__name__)

class Sniffer(Zigbee, EventsManager):
    """
    Zigbee Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device: Device):
        """Sniffer initialization.

        :param device: Device to use for sniffing
        :type device: WhadDevice
        """
        Zigbee.__init__(self, device)
        EventsManager.__init__(self)


        self.__configuration = SnifferConfiguration()
        self.__decryptor = ZigbeeDecryptor()
        self.__touchlink_key_derivation = TouchlinkKeyManager()
        self.__transport_key_distribution = TransportKeyDistribution()
        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)
        self.sniff_zigbee(channel=self.__configuration.channel)

    def add_key(self, key: bytes):
        """Add an encryption key to our sniffer.

        :param key: encryption key to add
        :type key: bytes
        """
        self.__configuration.keys.append(key)

    def clear_keys(self):
        """Clear all stored encryption keys.
        """
        self.__configuration.keys = []

    @property
    def decrypt(self) -> bool:
        """Decryption enabled
        """
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt: bool):
        """Set decryption status
        """
        self.__configuration.decrypt = decrypt


    @property
    def channel(self) -> int:
        """Current channel
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel: int = 11):
        """Set current channel.

        :param channel: new ZigBee channel to use
        :type channel: int
        """
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    @property
    def configuration(self) -> SnifferConfiguration:
        """Current sniffer configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def process_packet(self, packet: Packet):
        """Process received ZigBee packet.

        :param packet: received packet
        :type packet: :class:`scapy.packet.Packet`
        :return: received packet
        :rtype: :class:`scapy.packet.Packet`
        """
        if self.__touchlink_key_derivation.unencrypted_key is not None:
            logger.info("[i] New key extracted: %s",
                        self.__touchlink_key_derivation.unencrypted_key.hex())
            self.trigger_event(KeyExtractedEvent(self.__touchlink_key_derivation.unencrypted_key))
            self.add_key(self.__touchlink_key_derivation.unencrypted_key)
            self.__decryptor.add_key(self.__touchlink_key_derivation.unencrypted_key)
            self.__touchlink_key_derivation.reset()

        if self.__transport_key_distribution.transport_key is not None:
            logger.info(
                "[i] New key extracted: %s",
                self.__transport_key_distribution.transport_key.hex())
            self.trigger_event(KeyExtractedEvent(self.__transport_key_distribution.transport_key))
            self.add_key(self.__transport_key_distribution.transport_key)
            self.__decryptor.add_key(self.__transport_key_distribution.transport_key)
            self.__transport_key_distribution.reset()


        if ZigbeeSecurityHeader in packet and self.__configuration.decrypt:
            decrypted, success = self.__decryptor.attempt_to_decrypt(packet)
            if success:
                packet.data = decrypted
                packet.metadata.decrypted = True


        if self.__configuration.pairing:
            self.__touchlink_key_derivation.process_packet(packet)
            self.__transport_key_distribution.process_packet(packet)

        return packet

    def sniff(self, timeout: float = None) -> Generator[Packet, None, None]:
        """Sniff ZigBee packets from thin air.

        :param timeout: Specify the number of seconds after which sniffing will stop.
        :type timeout: float
        """
        # Save start time
        start = time()
        try:
            while True:
                if self.support_raw_pdu():
                    message_type = RawPduReceived
                else:
                    message_type = PduReceived

                message = self.wait_for_message(keep=message_filter(message_type), timeout=1.0)
                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    if packet is not None:
                        self.monitor_packet_rx(packet)
                        packet = self.process_packet(packet)
                        yield packet

                # Check if timeout has been reached (if provided)
                if timeout is not None:
                    if time() - start >= timeout:
                        break
        except WhadDeviceDisconnected:
            return



"""LoRaWAN 1.0 cryptographic primitives

This module provides multiple cryptographic primitives as specified
in the LoRaWAN 1.0 specifications, including various MIC computations,
key derivation and frame encryption/decryption.
"""
from struct import pack, unpack
from scapy.packet import Raw

from Cryptodome.Cipher import AES
from Cryptodome.Hash import CMAC

from whad.scapy.layers.lorawan import PHYPayload, MACPayloadDownlink, MACPayloadUplink
from .exceptions import BadMICError, MissingKeyError

def MIC(key : bytes, buffer : bytes) -> bytes:
    """Compute LoRaWAN MIC for PHY layer

    :param key: 128-bit key to use for MIC computation
    :type key: bytes
    :return: 
    """
    c = CMAC.new(key, ciphermod=AES)
    c.update(buffer)
    return c.digest()[:4]


def MIC_Uplink(key : bytes, dev_addr : int, fcnt : int, frame : bytes) -> bytes:
    """Compute uplink frame MIC

    :param key: Encryption key (depends on external conditions)
    :type key: bytes
    :param dev_addr: Device network address
    :type dev_addr: int
    :param fcnt: Current uplink frame counter
    :type fcnt: int
    :param frame: Frame used to generate the corresponding MIC
    :type frame: bytes
    :return: 4-byte message integrity code
    :rtype: bytes
    """
    b0 = pack('<BIBIIBB', 0x49, 0, 0, dev_addr, fcnt, 0, len(frame))
    return MIC(key, b0 + frame)

def MIC_Downlink(key : bytes, dev_addr : int, fcnt : int, frame : bytes) -> bytes:
    """Compute downlink frame MIC

    :param key: Encryption key (depends on external conditions)
    :type key: bytes
    :param dev_addr: Device network address
    :type dev_addr: int
    :param fcnt: Current downlink frame counter
    :type fcnt: int
    :param frame: Frame used to generate the corresponding MIC
    :type frame: bytes
    :return: 4-byte message integrity code
    :rtype: bytes
    """
    b0 = pack('<BIBIIBB', 0x49, 0, 1, dev_addr, fcnt, 0, len(frame))
    return MIC(key, b0 + frame)


def pad16(data : bytes) -> bytes:
    """Pad data to fit 16-byte multiple length.

    :param data: Data to pad
    :type data: bytes
    :return: 16-byte padded data
    :rtype: bytes
    """
    r = len(data)%16
    if r > 0:
        return data + b'\x00'*(16 - r)
    else:
        return data


def derive_nwkskey(appkey : bytes, join_nonce : int, netid : int, dev_nonce : int) -> bytes:
    """Derive NwkSKey from AppKey, Join and Dev nonces, Network ID

    :param appkey: Application Key
    :type appkey: bytes
    :param join_nonce: Join nonce
    :type join_nonce: int
    :param netid: Network ID
    :type netid: int
    :param dev_nonce: Device nonce
    :type dev_nonce: int
    :return: Derived network session encryption key
    :rtype: bytes
    """
    c = AES.new(appkey, mode=AES.MODE_ECB)
    buffer = b'\x01' + pack('<BHBHH', join_nonce&0xff, (join_nonce>>8)&0xffff, netid&0xff, (netid>>8)&0xffff , dev_nonce)
    return c.encrypt(pad16(buffer))



def derive_appskey(appkey : bytes, join_nonce : int, netid : int, dev_nonce : int) -> bytes:
    """Derive AppSKey from AppKey, Join and Dev nonces, Network ID

    :param appkey: Application Key
    :type appkey: bytes
    :param join_nonce: Join nonce
    :type join_nonce: int
    :param netid: Network ID
    :type netid: int
    :param dev_nonce: Device nonce
    :type dev_nonce: int
    :return: Derived application session key
    :rtype: bytes
    """
    c = AES.new(appkey, mode=AES.MODE_ECB)
    buffer = b'\x02' + pack('<BHBHH', join_nonce&0xff, (join_nonce>>8)&0xffff, netid&0xff, (netid>>8)&0xffff , dev_nonce)
    return c.encrypt(pad16(buffer))


def encrypt_frame(key : bytes, dev_addr : int, fcnt : int, frame : bytes, uplink : bool = True) -> bytes:
    """Encrypt a MAC frame with the provided key.

    :param key: Encryption/decryption key.
    :type key: bytes
    :param dev_addr: Device address
    :type dev_addr: int
    :param fcnt: Frame counter
    :type fcnt: int
    :param frame: Frame to encrypt or decrypt
    :type frame: bytes
    :param uplink: Uplink frame if True, downlink frame otherwise
    :type uplink: bool

    :return: Encrypted or decrypted frame
    :rtype: bytes
    """
    # Compute number of Ai blocks to generate
    frame_length = len(frame)
    nb_blocks = int(frame_length / 16)
    if frame_length % 16 > 0:
        nb_blocks += 1

    # Generate Ai blocks
    ai_blocks = []
    for i in range(nb_blocks):
        ai_blocks.append(pack('<BIBIIBB', 1, 0, 0 if uplink else 1, dev_addr,
                              fcnt, 0, i+1))
    
    # Generate our keystream
    c = AES.new(key,mode=AES.MODE_ECB)
    ks = b''
    for ai in ai_blocks:
        ks += c.encrypt(ai)

    # XOR our frame with our keystream
    output = b''
    for i in range(frame_length):
        output += bytes([frame[i] ^ ks[i]])
    
    # Return the result
    return output

def encrypt_fopts(key : bytes, dev_addr : int, fcnt : int, fopts : bytes, uplink : bool = True) -> bytes:
    """Encrypt a MAC frame with the provided key.

    :param key: Encryption/decryption key.
    :type key: bytes
    :param dev_addr: Device address
    :type dev_addr: int
    :param fcnt: Frame counter
    :type fcnt: int
    :param frame: Frame to encrypt or decrypt
    :type frame: bytes
    :param uplink: Uplink frame if True, downlink frame otherwise
    :type uplink: bool

    :return: Encrypted or decrypted frame options
    :rtype: bytes
    """

    # Generate A block
    ai_block = pack('<BIBIIBB', 1, 0, 0 if uplink else 1, dev_addr, fcnt, 0, 0)
    
    # Generate our keystream
    c = AES.new(key,mode=AES.MODE_ECB)
    ks = c.encrypt(ai_block)

    # XOR our frame with our keystream
    output = b''
    for i in range(len(fopts)):
        output += bytes([fopts[i] ^ ks[i]])
    
    # Return the result
    return output    

def decrypt_packet(packet : PHYPayload, appkey=None, appskey=None, nwkskey=None) -> PHYPayload:
    """Decrypt a LoRaWAN PHY payload given the provided keys.

    :param packet: LoRaWAN packet to decrypt
    :type packet: PHYPayload
    :param appkey: LoRaWAN Application Key
    :type appkey: bytes
    :param appskey: LoRaWAN Application Session Key
    :type appskey: bytes
    :param nwkskey: LoRaWAN Network Session Key

    :raises BadMICError: Incorrect MIC detected
    :raises MissingKeyError: A required encryption key is missing

    :return: decrypted LoRaWAN PHY packet
    :rtype: PHYPayload
    """
    if packet.mtype == 0x01:
        # Join Accept packet is encrypted with the appkey
        if appkey is not None:
            # First decrypt data
            phy_payload = bytes(packet)[1:]
            c = AES.new(appkey, mode=AES.MODE_ECB)
            dec_ja = c.encrypt(phy_payload)

            # Check MIC
            ja_data = bytes(packet)[0:1] + dec_ja[:-4]
            exp_mic = MIC(appkey, ja_data)
            mic = dec_ja[-4:]
            if exp_mic == mic:
                # MIC is ok, return decrypted packet
                return PHYPayload(ja_data + mic)
            else:
                raise BadMICError
        else:
            raise MissingKeyError('APPKey')
        
    elif packet.mtype == 0x02 or packet.mtype == 0x04:
        if nwkskey is not None and appskey is not None:
            # Decrypt uplink frame
            phy = bytes(packet)[:-4]
            mic = bytes(packet)[-4:]
            mac = packet.getlayer(MACPayloadUplink)
            exp_mic = MIC_Uplink(nwkskey, mac.dev_addr, mac.fcnt, phy)
            if exp_mic == mic:
                # decrypt mac commands
                dec_fopts = encrypt_fopts(appskey, mac.dev_addr, mac.fcnt, bytes(mac.fopts))
                mac.fopts = dec_fopts

                # decrypt payload
                if mac.fport == 0:
                    dec_payload = encrypt_frame(nwkskey, mac.dev_addr, mac.fcnt, bytes(mac.payload))
                else:
                    dec_payload = encrypt_frame(appskey, mac.dev_addr, mac.fcnt, bytes(mac.payload))
                    mac.payload = Raw(dec_payload)
                return packet
            else:
                raise BadMICError()
        else:
            if nwkskey is None:
                raise MissingKeyError('NwkSKey')
            if appskey is None:
                raise MissingKeyError('APPSKey')
    
    elif packet.mtype == 0x03 or packet.mtype == 0x05:
        if nwkskey is not None and appskey is not None:
            # Decrypt uplink frame
            phy = bytes(packet)[:-4]
            mic = bytes(packet)[-4:]
            mac = packet.getlayer(MACPayloadUplink)
            exp_mic = MIC_Downlink(nwkskey, mac.dev_addr, mac.fcnt, phy)
            if exp_mic == mic:
                # decrypt mac commands
                dec_fopts = encrypt_fopts(appskey, mac.dev_addr, mac.fcnt, bytes(mac.fopts), uplink=False)
                mac.fopts = dec_fopts
                
                if mac.fport == 0:
                    dec_payload = encrypt_frame(nwkskey, mac.dev_addr, mac.fcnt, bytes(mac.payload), uplink=False)
                else:
                    dec_payload = encrypt_frame(appskey, mac.dev_addr, mac.fcnt, bytes(mac.payload), uplink=False)
                mac.payload = Raw(dec_payload)
                return packet
            else:
                raise BadMICError
        else:
            if nwkskey is None:
                raise MissingKeyError('NwkSKey')
            if appskey is None:
                raise MissingKeyError('APPSKey')
    else:
        # not supported yet
        return packet

def encrypt_packet(packet : PHYPayload, appkey=None, appskey=None, nwkskey=None) -> PHYPayload:
    """Encrypt LoRaWAN packet.
    
    :param packet: LoRaWAN packet to decrypt
    :type packet: PHYPayload
    :param appkey: LoRaWAN Application Key
    :type appkey: bytes
    :param appskey: LoRaWAN Application Session Key
    :type appskey: bytes
    :param nwkskey: LoRaWAN Network Session Key

    :raises MissingKeyError: A required encryption key is missing

    :return: Encrypted LoRaWAN PHY packet
    :rtype: PHYPayload
    """
    if packet.mtype == 0x01:
        # Join Accept packet is encrypted with the appkey
        if appkey is not None:
            # Compute MIC before encrypting
            ja_data = bytes(packet)[:-4]
            mic = MIC(appkey, ja_data)

            # Encrypt join accept + mic
            ja_and_mic = bytes(packet)[1:-4] + mic
            c = AES.new(appkey, mode=AES.MODE_ECB)
            enc_ja = c.decrypt(ja_and_mic)
            enc_phy = b'\x20' + enc_ja

            # MIC is ok, return decrypted packet
            return PHYPayload(enc_phy)
        else:
            raise MissingKeyError('APPKey')
        
    elif packet.mtype == 0x02 or packet.mtype == 0x04:
        if nwkskey is not None and appskey is not None:
            # Encrypt uplink frame
            mac = packet.getlayer(MACPayloadUplink)

            # Encrypt Fopts
            enc_fopts = encrypt_fopts(appskey, mac.dev_addr, mac.fcnt, bytes(mac.fopts))
            mac.fopts = enc_fopts

            # Encrypt payload
            if mac.fport == 0:
                enc_payload = encrypt_frame(nwkskey, mac.dev_addr, mac.fcnt, bytes(mac.payload))
            else:
                enc_payload = encrypt_frame(appskey, mac.dev_addr, mac.fcnt, bytes(mac.payload))
            mac.payload = Raw(enc_payload)

            # Compute MIC
            phy = bytes(packet)[:-4]
            packet.mic = unpack('<I', MIC_Uplink(nwkskey, mac.dev_addr, mac.fcnt, phy))[0]
            return packet
        else:
            if nwkskey is None:
                raise MissingKeyError('NwkSKey')
            if appskey is None:
                raise MissingKeyError('APPSKey')
    elif packet.mtype == 0x03 or packet.mtype == 0x05:
        if nwkskey is not None and appskey is not None:
             # Encrypt downlink frame
            mac = packet.getlayer(MACPayloadDownlink)

            # Encrypt Fopts
            enc_fopts = encrypt_fopts(appskey, mac.dev_addr, mac.fcnt, bytes(mac.fopts), uplink=False)
            mac.fopts = enc_fopts

            # Encrypt payload
            if mac.fport == 0:
                enc_payload = encrypt_frame(nwkskey, mac.dev_addr, mac.fcnt, bytes(mac.payload), uplink=False)
            else:
                enc_payload = encrypt_frame(appskey, mac.dev_addr, mac.fcnt, bytes(mac.payload), uplink=False)
            mac.payload = Raw(enc_payload)

            # Compute MIC
            phy = bytes(packet)[:-4]
            packet.mic = unpack('<I', MIC_Downlink(nwkskey, mac.dev_addr, mac.fcnt, phy))[0]
            return packet
        else:
            if nwkskey is None:
                raise MissingKeyError('NwkSKey')
            if appskey is None:
                raise MissingKeyError('APPSKey')
    else:
        # not supported yet
        return packet


"""RF4CE Controller
"""
import logging

from whad.exceptions import UnsupportedCapability
from whad.dot15d4.stack import Dot15d4Stack
from whad.dot15d4.stack.mac import MACManager
from whad.dot15d4.address import Dot15d4Address

from .base import RF4CE
from ..stack.nwk import NWKManager


logger = logging.getLogger(__name__)

class Controller(RF4CE):
    """
    RF4CE Controller Node interface for compatible WHAD device.
    """
    def __init__(self, device, profiles=[]):
        RF4CE.__init__(self, device)

        # Check if device can act as target node

        if not self.can_be_end_device():
            raise UnsupportedCapability("Controller")

        # Stack initialization
        MACManager.add(NWKManager)
        self.__stack = Dot15d4Stack(self)

        self.__stack.set_short_address(0x1234)
        self.__stack.get_layer('mac').database.set('macShortAddress', 0x1234)
        self.__stack.get_layer('nwk').database.set("nwkcNodeCapabilities",
            (1 << 3) | # channel_normalization_capable
            (1 << 2) | # security_capable
            (0 << 1) | # power_source
            (0) # node_type
            #self.__stack.get_layer('nwk').database.get("nwkcNodeCapabilities") & 0xFE
        )

        # Channel initialization
        self.__channel = 15
        self.__channel_page = 0

        for profile in profiles:
            self.__stack.get_layer('apl').add_profile(profile)
            
        self.enable_reception()


    @property
    def stack(self):
        return self.__stack

    def enable_reception(self):
        self.set_end_device_mode(channel=self.__channel)

    def set_channel(self, channel=11):
        self.__channel = channel
        self.enable_reception()

    def perform_ed_scan(self, channel):
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        super().perform_ed_scan(channel)

    def set_channel_page(self, page=0):
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page


    def discovery(self):
        return self.__stack.get_layer('nwk').get_service('management').discovery()


    def get_channel(self):
        return self.__channel

    def get_channel_page(self):
        return self.__channel_page


    def send(self, packet):
        super().send(packet, channel=self.__channel)

    def on_pdu(self, pdu):
        if (
            hasattr(pdu,"metadata") and
            hasattr(pdu.metadata, "is_fcs_valid") and
            not pdu.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(pdu)

    def on_ed_sample(self, timestamp, sample):
        self.__stack.on_ed_sample(timestamp, sample)



"""WHAD Protocol Dot15d4 pdu messages abstraction layer.
"""
import logging

from struct import pack, unpack, error as StructError

from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS

from whad.scapy.layers.dot15d4tap import Dot15d4Raw
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldBool, dissect_failsafe
from whad.hub.dot15d4 import Dot15d4Domain, Dot15d4Metadata

logger = logging.getLogger(__name__)

@pb_bind(Dot15d4Domain, 'send', 1)
class SendPdu(PbMessageWrapper):
    """Send Dot15d4 PDU message class
    """
    channel = PbFieldInt('dot15d4.send.channel')
    pdu = PbFieldBytes('dot15d4.send.pdu')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        return Dot15d4(self.pdu)

    @staticmethod
    def from_packet(packet, channel: int = 11):
        """Convert a scapy packet to a SendPdu message
        """
        if Dot15d4 in packet:
            pdu = bytes(packet[Dot15d4])
        elif Dot15d4FCS in packet:
            pdu = bytes(packet[Dot15d4FCS])[:-2]
        else:
            return None

        msg = SendPdu(
            channel=channel,
            pdu=pdu
        )
        return msg

@pb_bind(Dot15d4Domain, 'send_raw', 1)
class SendRawPdu(PbMessageWrapper):
    """Send Dot15d4 raw PDU message class
    """
    channel = PbFieldInt('dot15d4.send_raw.channel')
    pdu = PbFieldBytes('dot15d4.send_raw.pdu')
    fcs = PbFieldInt('dot15d4.send_raw.fcs')

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding scapy packet
        """
        return Dot15d4FCS(self.pdu + bytes(pack('<H', self.fcs)))

    @staticmethod
    def from_packet(packet, channel: int = 11):
        """Convert a scapy packet to a SendPdu message
        """
        if Dot15d4FCS in packet:
            pdu = bytes(packet[Dot15d4FCS])[:-2]

            msg = SendRawPdu(
                channel=channel,
                pdu=pdu,
                fcs=packet.fcs
            )
        elif Dot15d4 in packet:
            # Convert packet to bytes
            pdu = bytes(packet[Dot15d4])

            # If packet len is greater than 2, truncate payload in order to
            # put the remaining 2 bytes in its FCS field
            if len(pdu) > 2:
                fcs = unpack("<H", pdu[-2:])[0]
                pdu = pdu[:-2]
            else:
                fcs = 0

            msg = SendRawPdu(
                channel=channel,
                pdu=pdu,
                fcs=fcs
            )
        elif Dot15d4Raw in packet:
            pdu = bytes(packet)

            # If packet len is greater than 2, truncate payload in order to
            # put the remaining 2 bytes in its FCS field
            if len(pdu) > 2:
                fcs = unpack("<H", pdu[-2:])[0]
                pdu = pdu[:-2]
            else:
                fcs = 0

            msg = SendRawPdu(
                channel=channel,
                pdu=pdu,
                fcs=fcs
            )
        else:
            return None


        return msg

@pb_bind(Dot15d4Domain, 'pdu', 1)
class PduReceived(PbMessageWrapper):
    """Dot15d4 PDU received message class
    """
    channel = PbFieldInt('dot15d4.pdu.channel')
    pdu = PbFieldBytes('dot15d4.pdu.pdu')
    rssi = PbFieldInt('dot15d4.pdu.rssi', optional=True)
    timestamp = PbFieldInt('dot15d4.pdu.timestamp', optional=True)
    fcs_validity = PbFieldBool('dot15d4.pdu.fcs_validity', optional=True)
    lqi = PbFieldInt('dot15d4.pdu.lqi', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to its scapy packet representation
        """
        # Create packet
        packet = Dot15d4(bytes(self.pdu))

        # Set packet metadata
        packet.metadata = Dot15d4Metadata()
        packet.metadata.channel = self.channel

        packet.metadata.decrypted = False

        if self.lqi is not None:
            packet.metadata.lqi = self.lqi
        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.fcs_validity is not None:
            packet.metadata.is_fcs_valid = self.fcs_validity

        # Return packet
        return packet

    @staticmethod
    def from_packet(packet):
        """Convert scapy packet to a PduReceived message
        """
        # Create a PduReceived message
        msg = PduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet.getlayer(Dot15d4)),
        )
        # Add optional metadata

        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.lqi is not None:
            msg.lqi = packet.metadata.lqi
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.rssi = packet.metadata.timestamp
        if packet.metadata.is_fcs_valid is not None:
            msg.fcs_validity = packet.metadata.is_fcs_valid

        # Return metadata
        return msg


@pb_bind(Dot15d4Domain, 'raw_pdu', 1)
class RawPduReceived(PbMessageWrapper):
    """Dot15d4 raw PDU received message class
    """
    channel = PbFieldInt('dot15d4.raw_pdu.channel')
    pdu = PbFieldBytes('dot15d4.raw_pdu.pdu')
    fcs = PbFieldInt('dot15d4.raw_pdu.fcs')
    rssi = PbFieldInt('dot15d4.raw_pdu.rssi', optional=True)
    timestamp = PbFieldInt('dot15d4.raw_pdu.timestamp', optional=True)
    fcs_validity = PbFieldBool('dot15d4.raw_pdu.fcs_validity', optional=True)
    lqi = PbFieldInt('dot15d4.raw_pdu.lqi', optional=True)

    @dissect_failsafe
    def to_packet(self):
        """Convert message to scapy packet.
        """
        try:
            # Create packet
            #print('converting %s' % (self.pdu + bytes(pack(">H", self.fcs))).hex())
            packet = Dot15d4FCS(bytes(self.pdu) + bytes(pack("<H", self.fcs)))

            # Set packet metadata
            packet.metadata = Dot15d4Metadata()
            packet.metadata.channel = self.channel
            packet.metadata.decrypted = False
            if self.lqi is not None:
                packet.metadata.lqi = self.lqi
            if self.rssi is not None:
                packet.metadata.rssi = self.rssi
            if self.timestamp is not None:
                packet.metadata.timestamp = self.timestamp
            if self.fcs_validity is not None:
                packet.metadata.is_fcs_valid = self.fcs_validity

            return packet
        except StructError:
            #logger.debug("[hub::dot15d4] error parsing 802.15.4 frame (%s)",
            #             bytes(self.pdu).hex() + bytes(pack("<H", self.fcs)).hex())
            logger.debug("[hub::dot15d4] considering 802.15.4 as raw frame")

            # Build packet
            packet = Dot15d4Raw(bytes(self.pdu) + pack("<H", self.fcs))

            # Set packet metadata
            packet.metadata = Dot15d4Metadata()
            packet.metadata.channel = self.channel

            packet.metadata.decrypted = False

            if self.lqi is not None:
                packet.metadata.lqi = self.lqi
            if self.rssi is not None:
                packet.metadata.rssi = self.rssi
            if self.timestamp is not None:
                packet.metadata.timestamp = self.timestamp
            if self.fcs_validity is not None:
                packet.metadata.is_fcs_valid = self.fcs_validity

            return packet 

    @staticmethod
    def from_packet(packet):
        """Convert packet to a RawPduReceived message.
        """
        # Create a PduReceived message
        msg = RawPduReceived(
            channel=packet.metadata.channel,
            pdu=bytes(packet.getlayer(Dot15d4FCS))[:-2],
            fcs=packet.fcs
        )

        # Add optional metadata
        if packet.metadata.decrypted is not None:
            msg.decrypted = packet.metadata.decrypted
        if packet.metadata.lqi is not None:
            msg.lqi = packet.metadata.lqi
        if packet.metadata.rssi is not None:
            msg.rssi = packet.metadata.rssi
        if packet.metadata.timestamp is not None:
            msg.timestamp = packet.metadata.timestamp
        if packet.metadata.is_fcs_valid is not None:
            msg.fcs_validity = packet.metadata.is_fcs_valid

        # Return metadata
        return msg



from scapy.layers.zigbee import ZigbeeDeviceProfile
from scapy.fields import BitField, XLEShortField, ByteEnumField, \
    ByteField, ConditionalField, FieldListField, FieldLenField
from scapy.layers.dot15d4 import dot15d4AddressField
from scapy.packet import Packet

class ZDPActiveEPReq(Packet):
    name = "ZDP Transaction Data: Active_EP_req"
    fields_desc = [
        # NWK Address (2 octets)
        XLEShortField("nwk_addr", 0),
    ]

class ZDPActiveEPRsp(Packet):
    name = "ZDP Transaction Data: Active_EP_rsp"
    fields_desc = [
        ByteEnumField("status", 0, {0:"success", 1:"device_not_found", 2:"inv_requesttype", 3:"no_descriptor"}),
        # NWK Address (2 octets)
        XLEShortField("nwk_addr", 0),
        ConditionalField(FieldLenField("num_active_endpoints", None, length_of="endpoints", fmt="B"),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            FieldListField("active_endpoints",[],ByteField("", 0), length_from=lambda p:p.num_active_endpoints),
            lambda pkt:pkt.getfieldval("status") == 0
        ),

    ]

class ZDPDeviceAnnce(Packet):
    name = "ZDP Transaction Data: Device_annce"
    fields_desc = [
        # NWK Address (2 octets)
        XLEShortField("nwk_addr", 0),
        # IEEE Address (8 octets)
        dot15d4AddressField("ieee_addr", 0, adjust=lambda pkt, x: 8),
        # Capability Information (1 octet)
        BitField("allocate_address", 0, 1),
        BitField("security_capability", 0, 1),
        BitField("reserved2", 0, 1),
        BitField("reserved1", 0, 1),
        BitField("receiver_on_when_idle", 0, 1),
        BitField("power_source", 0, 1),
        BitField("device_type", 0, 1),
        BitField("alternate_pan_coordinator", 0, 1),
    ]

class ZDPNodeDescReq(Packet):
    name = "ZDP Transaction Data: Node_desc_req"
    fields_desc = [
        # NWK Address (2 octets)
        XLEShortField("nwk_addr", 0),
    ]


class ZDPNodeDescRsp(Packet):
    name = "ZDP Transaction Data: Node_desc_rsp"
    fields_desc = [
        ByteEnumField("status", 0, {0:"success", 1:"device_not_found", 2:"inv_requesttype", 3:"no_descriptor"}),
        # NWK Address (2 octets)
        XLEShortField("nwk_addr", 0),
        ConditionalField(
            BitField("reserved0", 0, 3),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("user_descriptor_available", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("complex_descriptor_available", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("logical_type", 0, 3),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved1bis", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("support_2400_mhz", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("support_902_mhz", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved1", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("support_868_mhz", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("aps_flags", 0, 3),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("allocate_address", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("security_capability", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved2", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved3", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("receiver_on_when_idle", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("power_source", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("device_type", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("alternate_pan_coordinator", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            XLEShortField("manufacturer_code", None),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            ByteField("max_buffer_size", None),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            XLEShortField("max_incoming_transfer_size", None),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_primary_trust_center", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_backup_trust_center", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_primary_binding_table_cache", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_backup_binding_table_cache", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_primary_discovery_cache", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("server_backup_discovery_cache", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("network_manager", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved4", 0, 2),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("stack_compliance_revision", 0, 7),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            XLEShortField("max_outgoing_transfer_size", None),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("extended_active_endpoint_list_available", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("extended_simple_descriptors_list_available", 0, 1),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            BitField("reserved5", 0, 6),
            lambda pkt:pkt.getfieldval("status") == 0
        )
    ]

class ZDPNWKAddrReq(Packet):
    name = "ZDP Transaction Data: NWK_addr_req"
    fields_desc = [
        # IEEE Address (8 octets)
        dot15d4AddressField("ieee_addr", 0, adjust=lambda pkt, x: 8),
        ByteEnumField("request_type", 0, {0:"single_device_response", 1:"extended_response"}),
        ByteField("start_index", 0)
    ]

class ZDPNWKAddrRsp(Packet):
    name = "ZDP Transaction Data: NWK_addr_rsp"
    fields_desc = [
        ByteEnumField("status", 0, {0:"success", 1:"device_not_found", 2:"inv_requesttype", 3:"no_descriptor"}),
        dot15d4AddressField("ieee_addr", 0, adjust=lambda pkt, x: 8),
        XLEShortField("nwk_addr", 0),
        ConditionalField(
            FieldLenField("num_assoc_dev", None, length_of="associated_devices"),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(ByteField("start_index", 0x0),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            FieldListField("associated_devices",[],XLEShortField("", 0), length_from=lambda p:p.num_assoc_dev),
            lambda pkt:pkt.getfieldval("status") == 0
        ),

    ]

class ZDPIEEEAddrReq(Packet):
    name = "ZDP Transaction Data: IEEE_addr_req"
    fields_desc = [
        # IEEE Address (8 octets)
        XLEShortField("nwk_addr", 0),
        ByteEnumField("request_type", 0, {0:"single_device_response", 1:"extended_response"}),
        ByteField("start_index", 0)
    ]

class ZDPIEEEAddrRsp(Packet):
    name = "ZDP Transaction Data: IEEE_addr_rsp"
    fields_desc = [
        ByteEnumField("status", 0, {0:"success", 1:"device_not_found", 2:"inv_requesttype", 3:"no_descriptor"}),
        dot15d4AddressField("ieee_addr", 0, adjust=lambda pkt, x: 8),
        XLEShortField("nwk_addr", 0),
        ConditionalField(ByteField("num_assoc_dev", None),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(ByteField("start_index", 0x0),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            FieldListField("associated_devices",[],XLEShortField("", 0), length_from=lambda p:p.num_assoc_dev * 2),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
    ]

class ZDPSimpleDescReq(Packet):
    name = "ZDP Transaction Data: Simple_Desc_Req"
    fields_desc = [
        # IEEE Address (8 octets)
        XLEShortField("nwk_addr", 0),
        ByteField("endpoint", 0)
    ]

class ZDPSimpleDescRsp(Packet):
    name = "ZDP Transaction Data: Simple_Desc_Rsp"
    fields_desc = [
        ByteEnumField("status", 0, {0:"success", 1:"device_not_found", 2:"inv_requesttype", 3:"no_descriptor"}),
        # IEEE Address (8 octets)
        XLEShortField("nwk_addr", 0),
        ByteField("descriptor_length", 0),
        ConditionalField(ByteField("endpoint", 0x0),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(XLEShortField("profile_identifier", 0x0),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(XLEShortField("device_identifier", 0x0),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(BitField("device_version", 0 , 4),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(BitField("reserved", 0 , 4),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(FieldLenField("input_clusters_count", None, length_of="input_clusters", fmt="B"),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            FieldListField("input_clusters",[],XLEShortField("", 0), length_from=lambda p:p.input_clusters_count * 2),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(FieldLenField("output_clusters_count", None, length_of="output_clusters", fmt="B"),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
        ConditionalField(
            FieldListField("output_clusters",[],XLEShortField("", 0), length_from=lambda p:p.output_clusters_count * 2),
            lambda pkt:pkt.getfieldval("status") == 0
        ),
    ]

def guess_payload_class(self, payload):
    if self.underlayer.cluster == 0x0000:
        return ZDPNWKAddrReq
    elif self.underlayer.cluster == 0x0001:
        return ZDPIEEEAddrReq
    elif self.underlayer.cluster == 0x0002:
        return ZDPNodeDescReq
    elif self.underlayer.cluster == 0x0004:
        return ZDPSimpleDescReq
    elif self.underlayer.cluster == 0x0005:
        return ZDPActiveEPReq
    elif self.underlayer.cluster == 0x8000:
        return ZDPNWKAddrRsp
    elif self.underlayer.cluster == 0x8001:
        return ZDPIEEEAddrRsp
    elif self.underlayer.cluster == 0x8002:
        return ZDPNodeDescRsp
    elif self.underlayer.cluster == 0x8004:
        return ZDPSimpleDescRsp
    elif self.underlayer.cluster == 0x8005:
        return ZDPActiveEPRsp
    elif self.underlayer.cluster == 0x0013:
        return ZDPDeviceAnnce
    return Packet.guess_payload_class(self, payload)

ZigbeeDeviceProfile.guess_payload_class = guess_payload_class



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/phy/phy.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/phy/phy.proto\x12\x03phy\"\"\n\x13SetASKModulationCmd\x12\x0b\n\x03ook\x18\x01 \x01(\x08\"(\n\x13SetFSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\")\n\x14Set4FSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\")\n\x14SetGFSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\"(\n\x13SetMSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\"\x16\n\x14SetBPSKModulationCmd\"+\n\x14SetQPSKModulationCmd\x12\x13\n\x0boffset_qpsk\x18\x01 \x01(\x08\"\xde\x01\n\x14SetLoRaModulationCmd\x12\x11\n\tbandwidth\x18\x01 \x01(\r\x12\x32\n\x10spreading_factor\x18\x02 \x01(\x0e\x32\x18.phy.LoRaSpreadingFactor\x12(\n\x0b\x63oding_rate\x18\x03 \x01(\x0e\x32\x13.phy.LoRaCodingRate\x12\x17\n\x0fpreamble_length\x18\x04 \x01(\r\x12\x12\n\nenable_crc\x18\x05 \x01(\x08\x12\x15\n\rexplicit_mode\x18\x06 \x01(\x08\x12\x11\n\tinvert_iq\x18\x07 \x01(\x08\"\x1c\n\x1aGetSupportedFrequenciesCmd\"$\n\x0fSetFrequencyCmd\x12\x11\n\tfrequency\x18\x01 \x01(\r\"\x1e\n\x0eSetDataRateCmd\x12\x0c\n\x04rate\x18\x01 \x01(\r\"7\n\x10SetEndiannessCmd\x12#\n\nendianness\x18\x01 \x01(\x0e\x32\x0f.phy.Endianness\"/\n\rSetTXPowerCmd\x12\x1e\n\x08tx_power\x18\x01 \x01(\x0e\x32\x0c.phy.TXPower\"\'\n\x10SetPacketSizeCmd\x12\x13\n\x0bpacket_size\x18\x01 \x01(\r\"#\n\x0eSetSyncWordCmd\x12\x11\n\tsync_word\x18\x01 \x01(\x0c\"0\n\x08SniffCmd\x12\x16\n\tiq_stream\x18\x01 \x01(\x08H\x00\x88\x01\x01\x42\x0c\n\n_iq_stream\"\x19\n\x07SendCmd\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\"\x1c\n\nSendRawCmd\x12\x0e\n\x02iq\x18\x01 \x03(\x05\x42\x02\x10\x01\"4\n\x0fScheduleSendCmd\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"<\n\x12SchedulePacketResp\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x11\n\x04\x66ull\x18\x02 \x01(\x08H\x00\x88\x01\x01\x42\x07\n\x05_full\" \n\x12SchedulePacketSent\x12\n\n\x02id\x18\x01 \x01(\x05\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"(\n\x06JamCmd\x12\x1e\n\x04mode\x18\x01 \x01(\x0e\x32\x10.phy.JammingMode\"\x0c\n\nMonitorCmd\"\xf2\x01\n\x0ePacketReceived\x12\x11\n\tfrequency\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x0e\n\x06packet\x18\x04 \x01(\x0c\x12\x11\n\tdeviation\x18\x05 \x01(\r\x12\x10\n\x08\x64\x61tarate\x18\x06 \x01(\r\x12\x1f\n\x06\x65ndian\x18\x07 \x01(\x0e\x32\x0f.phy.Endianness\x12#\n\nmodulation\x18\x08 \x01(\x0e\x32\x0f.phy.Modulation\x12\x10\n\x08syncword\x18\t \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestamp\"\x85\x02\n\x11RawPacketReceived\x12\x11\n\tfrequency\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x0e\n\x06packet\x18\x04 \x01(\x0c\x12\x0e\n\x02iq\x18\x05 \x03(\x05\x42\x02\x10\x01\x12\x11\n\tdeviation\x18\x06 \x01(\r\x12\x10\n\x08\x64\x61tarate\x18\x07 \x01(\r\x12\x1f\n\x06\x65ndian\x18\x08 \x01(\x0e\x32\x0f.phy.Endianness\x12#\n\nmodulation\x18\t \x01(\x0e\x32\x0f.phy.Modulation\x12\x10\n\x08syncword\x18\n \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestamp\"\x1f\n\nPacketSent\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"9\n\x10MonitoringReport\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x12\n\x06report\x18\x02 \x03(\rB\x02\x10\x01\"\x90\x01\n\x18SupportedFrequencyRanges\x12\x46\n\x10\x66requency_ranges\x18\x01 \x03(\x0b\x32,.phy.SupportedFrequencyRanges.FrequencyRange\x1a,\n\x0e\x46requencyRange\x12\r\n\x05start\x18\x01 \x01(\r\x12\x0b\n\x03\x65nd\x18\x02 \x01(\r\"\x91\n\n\x07Message\x12+\n\x07mod_ask\x18\x01 \x01(\x0b\x32\x18.phy.SetASKModulationCmdH\x00\x12+\n\x07mod_fsk\x18\x02 \x01(\x0b\x32\x18.phy.SetFSKModulationCmdH\x00\x12-\n\x08mod_gfsk\x18\x03 \x01(\x0b\x32\x19.phy.SetGFSKModulationCmdH\x00\x12-\n\x08mod_bpsk\x18\x04 \x01(\x0b\x32\x19.phy.SetBPSKModulationCmdH\x00\x12-\n\x08mod_qpsk\x18\x05 \x01(\x0b\x32\x19.phy.SetQPSKModulationCmdH\x00\x12-\n\x08mod_4fsk\x18\x06 \x01(\x0b\x32\x19.phy.Set4FSKModulationCmdH\x00\x12+\n\x07mod_msk\x18\x07 \x01(\x0b\x32\x18.phy.SetMSKModulationCmdH\x00\x12=\n\x12get_supported_freq\x18\x08 \x01(\x0b\x32\x1f.phy.GetSupportedFrequenciesCmdH\x00\x12(\n\x08set_freq\x18\t \x01(\x0b\x32\x14.phy.SetFrequencyCmdH\x00\x12\'\n\x08\x64\x61tarate\x18\n \x01(\x0b\x32\x13.phy.SetDataRateCmdH\x00\x12+\n\nendianness\x18\x0b \x01(\x0b\x32\x15.phy.SetEndiannessCmdH\x00\x12&\n\x08tx_power\x18\x0c \x01(\x0b\x32\x12.phy.SetTXPowerCmdH\x00\x12,\n\x0bpacket_size\x18\r \x01(\x0b\x32\x15.phy.SetPacketSizeCmdH\x00\x12(\n\tsync_word\x18\x0e \x01(\x0b\x32\x13.phy.SetSyncWordCmdH\x00\x12\x1e\n\x05sniff\x18\x0f \x01(\x0b\x32\r.phy.SniffCmdH\x00\x12\x1c\n\x04send\x18\x10 \x01(\x0b\x32\x0c.phy.SendCmdH\x00\x12#\n\x08send_raw\x18\x11 \x01(\x0b\x32\x0f.phy.SendRawCmdH\x00\x12\x1e\n\x05start\x18\x12 \x01(\x0b\x32\r.phy.StartCmdH\x00\x12\x1c\n\x04stop\x18\x13 \x01(\x0b\x32\x0c.phy.StopCmdH\x00\x12\x1a\n\x03jam\x18\x14 \x01(\x0b\x32\x0b.phy.JamCmdH\x00\x12\"\n\x07monitor\x18\x15 \x01(\x0b\x32\x0f.phy.MonitorCmdH\x00\x12%\n\x06packet\x18\x16 \x01(\x0b\x32\x13.phy.PacketReceivedH\x00\x12,\n\nraw_packet\x18\x17 \x01(\x0b\x32\x16.phy.RawPacketReceivedH\x00\x12\x1d\n\x06jammed\x18\x18 \x01(\x0b\x32\x0b.phy.JammedH\x00\x12/\n\x0emonitor_report\x18\x19 \x01(\x0b\x32\x15.phy.MonitoringReportH\x00\x12\x37\n\x0esupported_freq\x18\x1a \x01(\x0b\x32\x1d.phy.SupportedFrequencyRangesH\x00\x12-\n\x08mod_lora\x18\x1b \x01(\x0b\x32\x19.phy.SetLoRaModulationCmdH\x00\x12*\n\nsched_send\x18\x1c \x01(\x0b\x32\x14.phy.ScheduleSendCmdH\x00\x12\x30\n\rsched_pkt_rsp\x18\x1d \x01(\x0b\x32\x17.phy.SchedulePacketRespH\x00\x12\x31\n\x0esched_pkt_sent\x18\x1e \x01(\x0b\x32\x17.phy.SchedulePacketSentH\x00\x42\x05\n\x03msg*\xa7\x03\n\nPhyCommand\x12\x14\n\x10SetASKModulation\x10\x00\x12\x14\n\x10SetFSKModulation\x10\x01\x12\x15\n\x11SetGFSKModulation\x10\x02\x12\x15\n\x11SetBPSKModulation\x10\x03\x12\x15\n\x11SetQPSKModulation\x10\x04\x12\x15\n\x11Set4FSKModulation\x10\x05\x12\x14\n\x10SetMSKModulation\x10\x06\x12\x1b\n\x17GetSupportedFrequencies\x10\x07\x12\x10\n\x0cSetFrequency\x10\x08\x12\x0f\n\x0bSetDataRate\x10\t\x12\x11\n\rSetEndianness\x10\n\x12\x0e\n\nSetTXPower\x10\x0b\x12\x11\n\rSetPacketSize\x10\x0c\x12\x0f\n\x0bSetSyncWord\x10\r\x12\t\n\x05Sniff\x10\x0e\x12\x08\n\x04Send\x10\x0f\x12\x0b\n\x07SendRaw\x10\x10\x12\x07\n\x03Jam\x10\x11\x12\x0b\n\x07Monitor\x10\x12\x12\t\n\x05Start\x10\x13\x12\x08\n\x04Stop\x10\x14\x12\x15\n\x11SetLoRaModulation\x10\x15\x12\x10\n\x0cScheduleSend\x10\x16*!\n\nEndianness\x12\x07\n\x03\x42IG\x10\x00\x12\n\n\x06LITTLE\x10\x01*(\n\x07TXPower\x12\x07\n\x03LOW\x10\x00\x12\n\n\x06MEDIUM\x10\x01\x12\x08\n\x04HIGH\x10\x02*+\n\x0bJammingMode\x12\x0e\n\nCONTINUOUS\x10\x00\x12\x0c\n\x08REACTIVE\x10\x01*N\n\x13LoRaSpreadingFactor\x12\x07\n\x03SF7\x10\x00\x12\x07\n\x03SF8\x10\x01\x12\x07\n\x03SF9\x10\x02\x12\x08\n\x04SF10\x10\x03\x12\x08\n\x04SF11\x10\x04\x12\x08\n\x04SF12\x10\x05*8\n\x0eLoRaCodingRate\x12\x08\n\x04\x43R45\x10\x00\x12\x08\n\x04\x43R46\x10\x01\x12\x08\n\x04\x43R47\x10\x02\x12\x08\n\x04\x43R48\x10\x03*\\\n\nModulation\x12\x07\n\x03\x41SK\x10\x00\x12\x07\n\x03\x46SK\x10\x01\x12\x0b\n\x07\x46OURFSK\x10\x02\x12\x08\n\x04GFSK\x10\x03\x12\x07\n\x03MSK\x10\x04\x12\x08\n\x04\x42PSK\x10\x05\x12\x08\n\x04QPSK\x10\x06\x12\x08\n\x04LORA\x10\x07\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.phy.phy_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SENDRAWCMD.fields_by_name['iq']._options = None
  _SENDRAWCMD.fields_by_name['iq']._serialized_options = b'\020\001'
  _RAWPACKETRECEIVED.fields_by_name['iq']._options = None
  _RAWPACKETRECEIVED.fields_by_name['iq']._serialized_options = b'\020\001'
  _MONITORINGREPORT.fields_by_name['report']._options = None
  _MONITORINGREPORT.fields_by_name['report']._serialized_options = b'\020\001'
  _PHYCOMMAND._serialized_start=3234
  _PHYCOMMAND._serialized_end=3657
  _ENDIANNESS._serialized_start=3659
  _ENDIANNESS._serialized_end=3692
  _TXPOWER._serialized_start=3694
  _TXPOWER._serialized_end=3734
  _JAMMINGMODE._serialized_start=3736
  _JAMMINGMODE._serialized_end=3779
  _LORASPREADINGFACTOR._serialized_start=3781
  _LORASPREADINGFACTOR._serialized_end=3859
  _LORACODINGRATE._serialized_start=3861
  _LORACODINGRATE._serialized_end=3917
  _MODULATION._serialized_start=3919
  _MODULATION._serialized_end=4011
  _SETASKMODULATIONCMD._serialized_start=36
  _SETASKMODULATIONCMD._serialized_end=70
  _SETFSKMODULATIONCMD._serialized_start=72
  _SETFSKMODULATIONCMD._serialized_end=112
  _SET4FSKMODULATIONCMD._serialized_start=114
  _SET4FSKMODULATIONCMD._serialized_end=155
  _SETGFSKMODULATIONCMD._serialized_start=157
  _SETGFSKMODULATIONCMD._serialized_end=198
  _SETMSKMODULATIONCMD._serialized_start=200
  _SETMSKMODULATIONCMD._serialized_end=240
  _SETBPSKMODULATIONCMD._serialized_start=242
  _SETBPSKMODULATIONCMD._serialized_end=264
  _SETQPSKMODULATIONCMD._serialized_start=266
  _SETQPSKMODULATIONCMD._serialized_end=309
  _SETLORAMODULATIONCMD._serialized_start=312
  _SETLORAMODULATIONCMD._serialized_end=534
  _GETSUPPORTEDFREQUENCIESCMD._serialized_start=536
  _GETSUPPORTEDFREQUENCIESCMD._serialized_end=564
  _SETFREQUENCYCMD._serialized_start=566
  _SETFREQUENCYCMD._serialized_end=602
  _SETDATARATECMD._serialized_start=604
  _SETDATARATECMD._serialized_end=634
  _SETENDIANNESSCMD._serialized_start=636
  _SETENDIANNESSCMD._serialized_end=691
  _SETTXPOWERCMD._serialized_start=693
  _SETTXPOWERCMD._serialized_end=740
  _SETPACKETSIZECMD._serialized_start=742
  _SETPACKETSIZECMD._serialized_end=781
  _SETSYNCWORDCMD._serialized_start=783
  _SETSYNCWORDCMD._serialized_end=818
  _SNIFFCMD._serialized_start=820
  _SNIFFCMD._serialized_end=868
  _SENDCMD._serialized_start=870
  _SENDCMD._serialized_end=895
  _SENDRAWCMD._serialized_start=897
  _SENDRAWCMD._serialized_end=925
  _SCHEDULESENDCMD._serialized_start=927
  _SCHEDULESENDCMD._serialized_end=979
  _SCHEDULEPACKETRESP._serialized_start=981
  _SCHEDULEPACKETRESP._serialized_end=1041
  _SCHEDULEPACKETSENT._serialized_start=1043
  _SCHEDULEPACKETSENT._serialized_end=1075
  _STARTCMD._serialized_start=1077
  _STARTCMD._serialized_end=1087
  _STOPCMD._serialized_start=1089
  _STOPCMD._serialized_end=1098
  _JAMCMD._serialized_start=1100
  _JAMCMD._serialized_end=1140
  _MONITORCMD._serialized_start=1142
  _MONITORCMD._serialized_end=1154
  _PACKETRECEIVED._serialized_start=1157
  _PACKETRECEIVED._serialized_end=1399
  _RAWPACKETRECEIVED._serialized_start=1402
  _RAWPACKETRECEIVED._serialized_end=1663
  _PACKETSENT._serialized_start=1665
  _PACKETSENT._serialized_end=1696
  _JAMMED._serialized_start=1698
  _JAMMED._serialized_end=1725
  _MONITORINGREPORT._serialized_start=1727
  _MONITORINGREPORT._serialized_end=1784
  _SUPPORTEDFREQUENCYRANGES._serialized_start=1787
  _SUPPORTEDFREQUENCYRANGES._serialized_end=1931
  _SUPPORTEDFREQUENCYRANGES_FREQUENCYRANGE._serialized_start=1887
  _SUPPORTEDFREQUENCYRANGES_FREQUENCYRANGE._serialized_end=1931
  _MESSAGE._serialized_start=1934
  _MESSAGE._serialized_end=3231
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/unifying/unifying.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%whad/protocol/unifying/unifying.proto\x12\x08unifying\"$\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"K\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0c\x12\x1d\n\x15show_acknowledgements\x18\x03 \x01(\x08\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"E\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"H\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"(\n\x15LogitechDongleModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"*\n\x17LogitechKeyboardModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\'\n\x14LogitechMouseModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"\x11\n\x0fSniffPairingCmd\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\xbe\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xbb\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xea\x04\n\x07Message\x12\x34\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x1b.unifying.SetNodeAddressCmdH\x00\x12#\n\x05sniff\x18\x02 \x01(\x0b\x32\x12.unifying.SniffCmdH\x00\x12\x1f\n\x03jam\x18\x03 \x01(\x0b\x32\x10.unifying.JamCmdH\x00\x12!\n\x04send\x18\x04 \x01(\x0b\x32\x11.unifying.SendCmdH\x00\x12(\n\x08send_raw\x18\x05 \x01(\x0b\x32\x14.unifying.SendRawCmdH\x00\x12\x31\n\x06\x64ongle\x18\x06 \x01(\x0b\x32\x1f.unifying.LogitechDongleModeCmdH\x00\x12\x35\n\x08keyboard\x18\x07 \x01(\x0b\x32!.unifying.LogitechKeyboardModeCmdH\x00\x12\x32\n\x05mouse\x18\x08 \x01(\x0b\x32!.unifying.LogitechKeyboardModeCmdH\x00\x12#\n\x05start\x18\t \x01(\x0b\x32\x12.unifying.StartCmdH\x00\x12!\n\x04stop\x18\n \x01(\x0b\x32\x11.unifying.StopCmdH\x00\x12\x32\n\rsniff_pairing\x18\x0e \x01(\x0b\x32\x19.unifying.SniffPairingCmdH\x00\x12\"\n\x06jammed\x18\x0b \x01(\x0b\x32\x10.unifying.JammedH\x00\x12+\n\x07raw_pdu\x18\x0c \x01(\x0b\x32\x18.unifying.RawPduReceivedH\x00\x12$\n\x03pdu\x18\r \x01(\x0b\x32\x15.unifying.PduReceivedH\x00\x42\x05\n\x03msg*\xc0\x01\n\x0fUnifyingCommand\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x08\n\x04Send\x10\x03\x12\x0b\n\x07SendRaw\x10\x04\x12\x16\n\x12LogitechDongleMode\x10\x05\x12\x18\n\x14LogitechKeyboardMode\x10\x06\x12\x15\n\x11LogitechMouseMode\x10\x07\x12\t\n\x05Start\x10\x08\x12\x08\n\x04Stop\x10\t\x12\x10\n\x0cSniffPairing\x10\nb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.unifying.unifying_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _UNIFYINGCOMMAND._serialized_start=1541
  _UNIFYINGCOMMAND._serialized_end=1733
  _SETNODEADDRESSCMD._serialized_start=51
  _SETNODEADDRESSCMD._serialized_end=87
  _SNIFFCMD._serialized_start=89
  _SNIFFCMD._serialized_end=164
  _JAMCMD._serialized_start=166
  _JAMCMD._serialized_end=191
  _SENDCMD._serialized_start=193
  _SENDCMD._serialized_end=262
  _SENDRAWCMD._serialized_start=264
  _SENDRAWCMD._serialized_end=336
  _LOGITECHDONGLEMODECMD._serialized_start=338
  _LOGITECHDONGLEMODECMD._serialized_end=378
  _LOGITECHKEYBOARDMODECMD._serialized_start=380
  _LOGITECHKEYBOARDMODECMD._serialized_end=422
  _LOGITECHMOUSEMODECMD._serialized_start=424
  _LOGITECHMOUSEMODECMD._serialized_end=463
  _STARTCMD._serialized_start=465
  _STARTCMD._serialized_end=475
  _STOPCMD._serialized_start=477
  _STOPCMD._serialized_end=486
  _SNIFFPAIRINGCMD._serialized_start=488
  _SNIFFPAIRINGCMD._serialized_end=505
  _JAMMED._serialized_start=507
  _JAMMED._serialized_end=534
  _RAWPDURECEIVED._serialized_start=537
  _RAWPDURECEIVED._serialized_end=727
  _PDURECEIVED._serialized_start=730
  _PDURECEIVED._serialized_end=917
  _MESSAGE._serialized_start=920
  _MESSAGE._serialized_end=1538
# @@protoc_insertion_point(module_scope)



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import sys
import logging
import time
import traceback
import string
from queue import Queue
from argparse import ArgumentParser

# Import every possible scapy layer for packet crafting
# pylint: disable-next=wildcard-import
from scapy.layers.all import *

# Import scapy helpers
from scapy.packet import Packet, Raw
from scapy.config import conf
from scapy.themes import BrightTheme

from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.cli.ui import warning, error, info, display_packet
from whad.cli.app import CommandLineApp, run_app
from whad.hub import ProtocolHub

# Import custom scapy layers required for packet crafting.
from whad.scapy.layers.rf4ce import *       # pylint: disable=wildcard-import,unused-wildcard-import
from whad.scapy.layers.phy import *         # pylint: disable=wildcard-import,unused-wildcard-import
from whad.scapy.layers.esb import *         # pylint: disable=wildcard-import,unused-wildcard-import
from whad.scapy.layers.unifying import *    # pylint: disable=wildcard-import,unused-wildcard-import
from whad.device.unix import UnixConnector
from whad.tools.utils import list_implemented_injectors, get_injector_parameters, \
    gen_option_name, build_configuration_from_args

logger = logging.getLogger(__name__)



class WhadDomainSubParser(ArgumentParser):
    """
    Implements a Whad Domain subparser.
    """
    def warning(self, message):
        """Display a warning message in orange (if color is enabled)
        """
        warning(message)

    def error(self, message):
        """Display an error message in red (if color is enabled)
        """
        error(message)
        #exit(1)

class WhadInjectApp(CommandLineApp):
    """winject main application class.
    """

    connector = None

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD generic injection tool',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

        self.add_argument(
            '-r',
            '--repeat',
            dest='repeat',
            action='store_true',
            default=False,
            help="Repeat the transmission of packets"
        )


        self.add_argument(
            '-d',
            '--delay',
            type=float,
            default='1.0',
            dest='delay',
            help="Delay between the transmission of two consecutive packets"
        )

        self.provided_count = 0
        self._input_queue = Queue()
        self._repeat_queue = Queue()

        # Load implemented injectors (introspection)
        self.environment = None

        # Build subparsers if stdin is not piped, when piped we will do
        # it in the pre_run() method.
        if not self.is_stdin_piped():
            self.subparsers = self.add_subparsers(
                required=True,
                dest="domain",
                parser_class=WhadDomainSubParser,
                help='Domain in use'
            )
            self.environment = list_implemented_injectors()
            self.build_subparsers(self.subparsers)
        else:
            self.subparsers = None

    def build_subparsers(self, subparsers):
        """
        Generate the subparsers argument according to the environment.
        """

        # Iterate over domain, and get the associated sniffer parameters
        for domain_name, domain in self.environment.items():
            domain["parameters"] = get_injector_parameters(
                domain["configuration_class"]
            )

            domain["subparser"] = subparsers.add_parser(
                domain_name,
                description=f"WHAD {domain_name.capitalize()} Injection tool"
            )

            # Iterate over every parameters, and add arguments to subparsers
            for (
                    parameter_name,
                    (parameter_type, parameter_default, parameter_base_class, parameter_help)
                ) in domain["parameters"].items():

                dest = parameter_name

                # If parameter is based on a dataclass, process a subparameter
                if parameter_base_class is not None:
                    _, parameter_name = parameter_name.split(".")

                parameter_name = gen_option_name(parameter_name)

                # Process parameter help and shortname
                if parameter_help is not None and "(" in parameter_help:
                    parameter_shortnames = [
                        f"-{i}" for i in
                        parameter_help.split("(")[1].replace(")","").split(",")
                    ]
                    parameter_help = parameter_help.split("(")[0]
                else:
                    parameter_shortnames = []

                # Process parameter type
                if parameter_type is not bool:
                    # If we got an int
                    if parameter_type is int:
                        # allow to provide hex arguments
                        # pylint: disable-next=unnecessary-lambda-assignment
                        parameter_type = lambda x: int(x,0)
                    # If we got a list, it is a list of string
                    if parameter_type is list:
                        parameter_default=[]
                        parameter_type=str
                        action = "append"
                    elif parameter_type is bytes:
                        parameter_type = bytes.fromhex
                        action = "store"
                    else:
                        action = "store"
                    # Add non-boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        default=parameter_default,
                        action=action,
                        type=parameter_type,
                        dest=dest,
                        help=parameter_help
                    )
                else:
                    # Add boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        action='store_true',
                        dest=dest,
                        help=parameter_help
                    )

            domain["subparser"].add_argument(
                dest='packet',
                nargs="*",
                help='packet to inject'
            )

    def on_incoming_packet(self, packet: Packet):
        """Incoming packet callback method, adds packet to input queue.
        """
        self._input_queue.put(packet)

    def generate_packets(self):
        """Convert packets provided by the user through command-line into scapy
        packets to inject.
        """
        self.provided_count = 0
        if hasattr(self.args, "packet"):
            for p in self.args.packet:
                # Is it some hex
                if all(c in string.hexdigits for c in p):
                    if len(p) % 2 == 0:
                        try:
                            pkt = bytes.fromhex(p)
                            self._input_queue.put(pkt)
                            self.provided_count += 1
                        except Exception as err:
                            error(f"Failure during raw packet decoding: {p} -> {err}")
                    else:
                        error("Raw packets must be provided in valid hexadecimal form")
                else:
                    try:
                        # pylint: disable-next=eval-used
                        pkt = eval(p)
                        self._input_queue.put(pkt)
                        self.provided_count += 1
                    except SyntaxError:
                        error(f"Invalid syntax: `{p}`")
                    except Exception as err:
                        error(f"Failure during packet interpretation: {p} -> {err}")

    def pre_run(self):
        """Pre-run operations: rewrite arguments to allow skipping domain name
        when a pipe is configured.
        """

        try:
            try:
                index = sys.argv.index("--interface")
            except ValueError:
                index = sys.argv.index("-i")
        except ValueError:
            if "-h" in sys.argv or "--help" in sys.argv:
                self.print_help()
                sys.exit(1)
            error("You need to provide an interface.")
            sys.exit(1)

        start_argv = sys.argv[:index + 2]
        end_argv =  sys.argv[index + 2:]

        if len(sys.argv) == 1 or "-h" in start_argv or "--help" in start_argv:
            self.print_help()
            sys.exit(1)

        error_func = self.error
        if self.subparsers is None:
            self.error  = lambda m : None

        # Load implemented injectors *before* calling super().__init__()
        # drastically improves performances... and I have no clue why o_o
        if self.environment is None:
            self.environment = list_implemented_injectors()

        # Call CLI app pre-run: it will initialize our source interface
        super().pre_run()

        try:
            domain = self.args.domain
        except AttributeError:
            self.error("You have to provide a domain.")
            sys.exit(1)

        if self.subparsers is None:
            self.error  = error_func

        if self.subparsers is None:
            sys.argv = start_argv

            self.subparsers = self.add_subparsers(
                required=True,
                dest="domain",
                parser_class=WhadDomainSubParser,
                help='Domain in use'
            )

            # We build subparsers.
            self.build_subparsers(self.subparsers)

            # And we inject the domain parameter retrieved from the parameters
            # given by the input tool through the Unix socket URI.
            if domain not in start_argv and domain not in end_argv:
                sys.argv = start_argv + [domain] + end_argv
            else:
                sys.argv = start_argv + end_argv

            # Parse winject arguments and propagate into the `self.args` namespace.
            for k, v in self.parse_args().__dict__.items():
                setattr(self.args, k, v)

        # If no color is not selected, configure scapy color theme
        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

        # Generate packets if provided through commandline
        self.generate_packets()


    def is_input_alive(self) -> bool:
        """Determine if the input interface is providing packets to inject.

        :return: `True` if input interface is alive, `False` otherwise.
        :rtype: bool
        """
        return self.input_interface is not None and self.input_interface.opened

    def has_pending_packets(self) -> bool:
        """Determine if we still have some packets to inject.

        :return: `True` if we have pending packets to inject, `False` otherwise.
        :rtype: bool
        """
        return self.provided_count > 0 or not self._input_queue.empty()

    def run(self):
        injector = None
        # Launch pre-run tasks
        self.pre_run()

        try:
            # We need to have an interface specified
            if self.interface is not None:
                # We need to have a domain specified
                if self.args.domain is not None:
                    # Parse the arguments to populate a injector configuration
                    configuration = build_configuration_from_args(self.environment, self.args)

                    # Generate an injector based on the selected domain
                    injector = self.environment[self.args.domain]["injector_class"](self.interface)


                    if self.is_piped_interface():
                        connector = UnixConnector(self.input_interface)

                        connector.domain = self.args.domain
                        hub = ProtocolHub(2)
                        connector.format = hub.get(self.args.domain).format
                        connector.on_packet = self.on_incoming_packet
                        connector.unlock()

                    try:
                        injector.configuration = configuration
                    except Exception as e:
                        self.error("Error during configuration: " +repr(e))

                    # Main loop
                    while True:

                        # Inject while input interface is alive or we have pending packets
                        while self.is_input_alive() or self.has_pending_packets():
                            if not self._input_queue.empty():
                                packet = self._input_queue.get()
                                try:
                                    if isinstance(packet, bytes):
                                        packet = injector.format(packet)[0]
                                        # patch for BLE only
                                        if BTLE_RF in packet:
                                            packet = packet[BTLE]
                                    info("Transmitting: ")
                                    display_packet(packet)
                                    injector.inject(packet)
                                    if float(self.args.delay) > 0:
                                        time.sleep(float(self.args.delay))
                                    if self.args.repeat:
                                        self._repeat_queue.put(packet)
                                except Exception as e:
                                    self.error("Error during injection: " + repr(e))
                                    traceback.print_exc()
                                self.provided_count-=1

                        # Break if no packets to repeat
                        if not self.args.repeat:
                            break

                        # Move packets to repeat into input queue (pending packets).
                        while not self._repeat_queue.empty():
                            self._input_queue.put(self._repeat_queue.get())
                            self.provided_count += 1
                else:
                    self.error("You need to specify a domain.")
            else:
                self.error("You need to specify an interface with option --interface.")

        except UnsupportedDomain:
            self.error(f"WHAD device doesn\'t support selected domain ({self.args.domain})")

        except UnsupportedCapability as unsupported_capability:
            self.error((f"WHAD device doesn't support selected capability "
                        f"({unsupported_capability.capability})"))

        except KeyboardInterrupt:
            self.warning("injector stopped (CTRL-C)")
            injector.stop()
            injector.close()
            sys.exit(1)


def winject_main():
    """Launcher for winject CLI application.
    """
    app = WhadInjectApp()
    run_app(app)



"""ATT constants (error and operation codes)
"""
import sys
from typing import List

if sys.version_info[0] == 3 and sys.version_info[1] >= 10:
    # Import UnionType only for Python >= 3.10
    from types import UnionType
    def is_union_type(obj):
        """Determine if object is of UnionType.
        """
        return isinstance(obj, UnionType)
else:
    # UnionType is not available
    def is_union_type(_):
        """UnionType is not available by default
        """
        return False

class BleAttOpcode:
    """ATT operation codes
    """
    ERROR_RESPONSE = 0x01
    EXCHANGE_MTU_REQUEST = 0x02
    EXCHANGE_MTU_RESPONSE = 0x03
    FIND_INFO_REQUEST = 0x04
    FIND_INFO_RESPONSE = 0x05
    FIND_BY_TYPE_VALUE_REQUEST = 0x06
    FIND_BY_TYPE_VALUE_RESPONSE = 0x07
    READ_BY_TYPE_REQUEST = 0x08
    READ_BY_TYPE_RESPONSE = 0x09
    READ_REQUEST = 0x0A
    READ_RESPONSE = 0x0B
    READ_BLOB_REQUEST = 0x0C
    READ_BLOB_RESPONSE = 0x0D
    READ_MULTIPLE_REQUEST = 0x0E
    READ_MULTIPLE_RESPONSE = 0x0F
    READ_BY_GROUP_TYPE_REQUEST = 0x10
    READ_BY_GROUP_TYPE_RESPONSE = 0x11
    WRITE_REQUEST = 0x12
    WRITE_RESPONSE = 0x13
    WRITE_COMMAND = 0x52
    SIGNED_WRITE_COMMAND = 0xD2
    PREPARE_WRITE_REQUEST = 0x16
    PREPARE_WRITE_RESPONSE = 0x17
    EXECUTE_WRITE_REQUEST = 0x18
    EXECUTE_WRITE_RESPONSE = 0x19
    HANDLE_VALUE_NOTIFICATION = 0x1B
    HANDLE_VALUE_INDICATION = 0x1D
    HANDLE_VALUE_CONFIRMATION = 0x1E

class BleAttErrorCode:
    """ATT error code
    """
    INVALID_HANDLE = 0x01
    READ_NOT_PERMITTED = 0x02
    WRITE_NOT_PERMITTED = 0x03
    INVALID_PDU = 0x04
    INSUFFICIENT_AUTHENT = 0x05
    REQUEST_NOT_SUPP = 0x06
    INVALID_OFFSET = 0x07
    INSUFFICIENT_AUTHOR = 0x08
    PREPARE_QUEUE_FULL = 0x09
    ATTRIBUTE_NOT_FOUND = 0x0A
    ATTRIBUTE_NOT_LONG = 0x0B
    INSUFFICIENT_ENC_KEY_SIZE = 0x0C
    INVALID_ATTR_VALUE_LENGTH = 0x0D
    UNLIKELY_ERROR = 0x0E
    INSUFFICIENT_ENCRYPTION = 0x0F
    UNSUPPORTED_GROUP_TYPE = 0x10
    INSUFFICIENT_RESOURCES = 0x11
    DB_OUT_OF_SYNC = 0x12
    VALUE_NOT_ALLOWED = 0x13

class SecurityMode:
    """SMP Security mode

    This class is used to store the security mode and level.
    """
    def __init__(self, security_mode=0, security_level=0):
        self.security_mode = security_mode
        self.security_level = security_level

class BleAttSecurityMode:
    """BLE default security modes
    """
    NO_ACCESS = SecurityMode(0, 0)
    OPEN = SecurityMode(1, 1)
    ENCRYPTION_NO_AUTHENTICATION = SecurityMode(1, 2)
    ENCRYPTION_WITH_AUTHENTICATION = SecurityMode(1, 3)
    ENCRYPTION_WITH_SECURE_CONNECTIONS = SecurityMode(1, 4)
    DATA_SIGNING_NO_AUTHENTICATION = SecurityMode(2, 1)
    DATA_SIGNING_WITH_AUTHENTICATION = SecurityMode(2, 2)

class BleAttProperties:
    """BLE Attribute properties
    """
    READ = 0x01
    WRITE = 0x02
    DEFAULT = READ | WRITE


class SecurityProperty:
    """BLE basic security property
    """
    def __repr__(self):
        return self.__class__.__name__

class Encryption(SecurityProperty):
    """Encryption security property
    """

class Authentication(SecurityProperty):
    """Authentication security property
    """

class Authorization(SecurityProperty):
    """Authorization security property
    """

class SecurityAccess:
    """Security access definition

    A security access is determined by one or more security properties.
    """
    TYPE = ""

    def __init__(self, *args):
        if len(args) > 0:
            if is_union_type(args[0]):
                self.__access = list(args[0].__args__)
            else:
                self.__access = list(args)
        else:
            self.__access = []

    def requires_encryption(self) -> bool:
        """Determine if encryption security property is set
        """
        return Encryption in self.__access

    def requires_authentication(self) -> bool:
        """Determine if authentication security property is set
        """
        return Authentication in self.__access

    def requires_authorization(self) -> bool:
        """Determine if authorization security property is set
        """
        return Authorization in self.__access

    @property
    def access(self):
        """Security access properties
        """
        return self.__access

    def __repr__(self):
        """String representation
        """
        props = " | ".join([a.__name__ for a in self.access])
        return f"{self.__class__.TYPE} ({props})"

    @classmethod
    def generate(cls, val):
        """Generate a security access based on a list of security properties.
        """
        if isinstance(val, list) and all([isinstance(i, SecurityAccess) for i in val]):
            return val

        if isinstance(val, SecurityAccess):
            return [val]

        return []

    def __or__(self, other):
        """overload the | operator to append a security property to the current
        properties.
        """
        if isinstance(other, SecurityAccess):
            return [self, other]

        return [self]

    @classmethod
    def accesses_to_int(cls, value) -> int:
        """Convert a security access into the corresponding integer value.
        """
        shift = 0
        return_value = 0
        for access in value:
            if isinstance(access, ReadAccess):
                shift = 0
            elif isinstance(access, WriteAccess):
                shift = 4

            if Encryption in access.access:
                return_value |= (1 << shift)
            if Authentication in access.access:
                return_value |= (2 << shift)
            if Authorization in access.access:
                return_value |= (4 << shift)
        return return_value

    @classmethod
    def int_to_accesses(cls, value) -> List[SecurityProperty]:
        """Convert an integer value into the corresponding security properties
        list
        """
        accesses = []
        read_properties = []
        if bool(value & 1):
            read_properties.append(Encryption)
        if bool(value & 2):
            read_properties.append(Authentication)
        if bool(value & 4):
            read_properties.append(Authorization)
        if len(read_properties) > 0:
            accesses.append(ReadAccess(*read_properties))

        value = value >> 4
        write_properties = []
        if bool(value & 1):
            write_properties.append(Encryption)
        if bool(value & 2):
            write_properties.append(Authentication)
        if bool(value & 4):
            write_properties.append(Authorization)
        if len(write_properties) > 0:
            accesses.append(WriteAccess(*write_properties))
        return accesses

class ReadAccess(SecurityAccess):
    """Default read access
    """
    TYPE = "Read"

class WriteAccess(SecurityAccess):
    """Default write access
    """
    TYPE = "Write"
